@@ -1,4 +1,4 @@
-/* Copyright (c) 1992, 1998, 2000 John E. Davis
+/* Copyright (c) 1992, 1998, 2000, 2005, 2006 John E. Davis
  * This file is part of JED editor library source.
  *
  * You may distribute this file under the terms the GNU General Public
@@ -11,19 +11,27 @@
 # include <stdlib.h>
 #endif
 
-#include <limits.h>
+#include <slang.h>
+#include "jedlimit.h"
 
-#ifdef PATH_MAX
-# define JED_MAX_PATH_LEN PATH_MAX
-#else
-# ifdef IBMPC_SYSTEM
-#  define JED_MAX_PATH_LEN 256
-# else
-#  define JED_MAX_PATH_LEN 1024
-# endif
+#if SLANG_VERSION >= 20000
+# define SLANG_V2
 #endif
 
-#include <slang.h>
+#ifdef SLANG_V2
+# define JED_HAS_UTF8_SUPPORT 1
+#else
+# define JED_HAS_UTF8_SUPPORT 0
+#endif
+
+#if JED_HAS_UTF8_SUPPORT
+# define JED_MAX_MULTIBYTE_SIZE SLUTF8_MAX_MBLEN
+extern int Jed_UTF8_Mode;
+#else
+typedef unsigned long SLwchar_Type;
+typedef unsigned char SLuchar_Type;
+# define JED_MAX_MULTIBYTE_SIZE 6
+#endif
 
 typedef struct _Buffer Buffer;
 
@@ -51,8 +59,8 @@ typedef struct Line
    int space;		       /* space allocated for line */
 #endif
 #if JED_HAS_LINE_ATTRIBUTES
-   /* The first 4 bits of the flags field represent a 4 bit integer that
-    * is used by the syntax parsing routines.  The next 4 bits
+   /* The first n bits of the flags field represent an n bit integer that
+    * is used by the syntax parsing routines.  The next m bits
     * control the various flags.  The last 8 bits specify the
     * color of the line.  
     *
@@ -66,16 +74,25 @@ typedef struct Line
 #define JED_LINE_IN_COMMENT		0x0001
 #define JED_LINE_IN_STRING0		0x0002
 #define JED_LINE_IN_STRING1		0x0004
-#define JED_LINE_HAS_EOL_COMMENT	0x0008
-#define JED_LINE_SYNTAX_BITS		0x000F
-   
-#define JED_LINE_HIDDEN			0x0010
-#define JED_LINE_IS_READONLY		0x0020
-   
-#define JED_LINE_COLOR_BITS		0xFF00
+#define JED_LINE_IN_HTML		0x0008
+#define JED_LINE_HAS_EOL_COMMENT	0x0010
+#define JED_LINE_SYNTAX_BITS		0x001F
+    
+#define JED_LINE_HIDDEN			0x0020
+#define JED_LINE_IS_READONLY		0x0040
+#define JED_LINE_FLAG_BITS		0x0060
 
+#define JED_LINE_COLOR_BITS		0xFF00
+#define JED_GET_LINE_COLOR(line) ((line->flags & JED_LINE_COLOR_BITS) >> 8)
+#define JED_SET_LINE_COLOR(line,color) \
+   (line)->flags &= ~JED_LINE_COLOR_BITS; \
+   (line)->flags |= (((color) << 8) & JED_LINE_COLOR_BITS)
 } Line;
 
+#define LINE_HAS_NEWLINE(l) \
+   ((l)->len && ((l)->data[(l)->len - 1] == '\n'))
+
+
 /* This is the price we pay for a linked list approach.  With straight
    buffer gap, this would be an integer.  Sigh. */
 typedef struct Mark
@@ -131,10 +148,38 @@ Jed_Save_Narrow_Type;
 typedef struct 
 {
    int tab;			       /* tab width */
+   int case_search;
+   int is_utf8;
 } Buffer_Local_Type;
 
 extern Buffer_Local_Type Buffer_Local;
 
+typedef struct
+{   
+   SLang_Name_Type *mark_paragraph_hook;
+   SLang_Name_Type *format_paragraph_hook;
+   SLang_Name_Type *forward_paragraph_hook;
+   SLang_Name_Type *backward_paragraph_hook;
+   SLang_Name_Type *par_sep;		       /* paragraph sep function */
+   SLang_Name_Type *indent_hook;
+   SLang_Name_Type *newline_indent_hook;
+   SLang_Name_Type *wrap_hook;
+   SLang_Name_Type *wrapok_hook;
+   SLang_Name_Type *bob_eob_error_hook;
+#ifdef HAS_MOUSE
+   SLang_Name_Type *mouse_down_hook;
+   SLang_Name_Type *mouse_up_hook;
+   SLang_Name_Type *mouse_drag_hook;
+# if JED_HAS_MULTICLICK
+   SLang_Name_Type *mouse_2click_hook;
+   SLang_Name_Type *mouse_3click_hook;
+# endif
+#endif
+   SLang_Name_Type *update_hook;
+   SLang_Name_Type *color_region_hook;
+}
+Jed_Buffer_Hook_Type;
+
 struct _Buffer
 {
    Line *beg;			       /* Top line of buffer */
@@ -146,10 +191,10 @@ struct _Buffer
    char *name;			       /* name of this buffer (slstring) */
    char *file;			       /* filename sans dir (slstring) */
    char *dir;			       /* directory of file (slstring) */
-   char *dirfile;		       /* canonical filename with expanded links */
+   char *dirfile;		       /* dir+file sans link expansion */
+   char *canonical_dirfile;	       /* canonical filename with expanded links */
 #ifdef REAL_UNIX_SYSTEM
-   int device;			       /* inode and device of DIRECTORY that file
-					* resides in */
+   int device;			       /* inode and device of canonical_dirfile */
    int inode;
 #endif
    int umask;
@@ -180,20 +225,8 @@ struct _Buffer
    Jed_Mark_Array_Type *mark_array;
    int vis_marks;		       /* number of visible marks */
    char status_line[80];
-   SLang_Name_Type *par_sep;		       /* paragraph sep function */
-   SLang_Name_Type *indent_hook;
-   SLang_Name_Type *newline_indent_hook;
-   SLang_Name_Type *wrap_hook;
-   SLang_Name_Type *bob_eob_error_hook;
-#ifdef HAS_MOUSE
-   SLang_Name_Type *mouse_down_hook;
-   SLang_Name_Type *mouse_up_hook;
-   SLang_Name_Type *mouse_drag_hook;
-# if JED_HAS_MULTICLICK
-   SLang_Name_Type *mouse_2click_hook;
-   SLang_Name_Type *mouse_3click_hook;
-# endif
-#endif
+   Jed_Buffer_Hook_Type *buffer_hooks;
+
 #if JED_HAS_COLOR_COLUMNS
    unsigned int coloring_style;
    unsigned char *column_colors;
@@ -220,42 +253,44 @@ struct _Buffer
 #if JED_HAS_TTY_MENUS
    Menu_Bar_Type *menubar;
 #endif
-   SLang_Name_Type *update_hook;
+#if JED_HAS_DISPLAY_LINE_NUMBERS
+   int line_num_display_size;
+#endif
 };
 
 extern char Default_Status_Line[80];
 
 /* flags */
-#define BUFFER_MODIFIED 0x01
+#define BUFFER_MODIFIED			0x0001
 
 /* This flag cannot be used with the AUTO_SAVE_JUST_SAVE flag */
-#define AUTO_SAVE_BUFFER 0x02
+#define AUTO_SAVE_BUFFER		0x0002
 /* these two flags are to tell user that the buffer and the file on disk
    have been modified--- see update_marks and main editor loop */
-#define FILE_MODIFIED 0x04
-#define READ_ONLY 0x08
-#define OVERWRITE_MODE 0x10
-#define UNDO_ENABLED 0x20
+#define FILE_MODIFIED			0x0004
+#define READ_ONLY			0x0008
+#define OVERWRITE_MODE			0x0010
+#define UNDO_ENABLED			0x0020
 
 /* skip this buffer if looking for a pop up one. */
-#define BURIED_BUFFER 0x40
+#define BURIED_BUFFER			0x0040
 
 /* Instead of autosaving saving the buffer, just save it.  This flag
  * is only used when SIGHUP or something like that hits.  It is also
  * used when exiting the editor.  It will cause the buffer to be silently 
  * saved.  It is possible that I need another flag for this.
  */
-#define AUTO_SAVE_JUST_SAVE 0x80
-#define NO_BACKUP_FLAG  0x100
-#define BINARY_FILE  0x200
-#define ADD_CR_ON_WRITE_FLAG 0x400
-
-#define ABBREV_MODE 0x800
+#define AUTO_SAVE_JUST_SAVE		0x0080
+#define NO_BACKUP_FLAG			0x0100
+#define BINARY_FILE			0x0200
+#define ADD_CR_ON_WRITE_FLAG		0x0400
+#define ABBREV_MODE			0x0800
+#define SMG_EMBEDDED_ESCAPE		0x1000
 
+#define BUFFER_NON_LOCKING		0x8000
 #ifndef VMS
-#define MAP_CR_TO_NL_FLAG 0x1000
+# define MAP_CR_TO_NL_FLAG		0x10000
 #endif
-#define BUFFER_NON_LOCKING	0x8000
 
 extern char *Read_Only_Error;
 extern char *Line_Read_Only_Error;
@@ -285,6 +320,26 @@ extern char *Line_Read_Only_Error;
 extern Buffer *CBuf;
 extern Line *CLine;
 
+extern void jed_set_point (int point);
+extern void jed_position_point (unsigned char *);
+extern unsigned char *jed_eol_position (Line *);
+
+extern unsigned char *
+  jed_multibyte_chars_forward (unsigned char *p, unsigned char *pmax,
+			       unsigned int n, unsigned int *dn, int skip_combining);
+extern unsigned char *
+  jed_multibyte_chars_backward (unsigned char *pmin, unsigned char *p,
+				unsigned int n, unsigned int *dn, int skip_combining);
+
+/* This is roughly equivalent to (UPPERCASE(*a++) - UPPERCASE(*b++)) */
+extern int jed_multibyte_charcasecmp (unsigned char **ap, unsigned char *amax,
+				      unsigned char **bp, unsigned char *bmax);
+
+/* Wide char equiv of *buf++ = c; */
+unsigned char *jed_wchar_to_multibyte (SLwchar_Type c, unsigned char *buf);
+
+extern int jed_what_char (SLwchar_Type *);
+extern void jed_count_chars (void);
 
 extern int bob(void);
 extern int eob(void);                  /* point to end of buffer */
@@ -296,11 +351,12 @@ extern int eobp(void);
 extern int eolp(void);
 extern int bolp(void);
 
-extern int prevline(int *);
-extern int nextline(int *);
+extern unsigned int jed_up (unsigned int);
+extern unsigned int jed_down (unsigned int);
 
-extern int forwchars(int *);
-extern int backwchars(int *);
+extern unsigned int jed_right (unsigned int);
+extern unsigned int jed_left (unsigned int);
+extern unsigned int jed_right_bytes (unsigned int);
 extern void goto_line(int *);
 
 extern Line *make_line1(unsigned int);
