@@ -1,4 +1,4 @@
-/* Copyright (c) 1992, 1998, 2000, 2002 John E. Davis
+/* Copyright (c) 1992, 1998, 2000, 2002, 2003, 2004, 2005, 2006 John E. Davis
  * This file is part of JED editor library source.
  *
  * You may distribute this file under the terms the GNU General Public
@@ -61,11 +61,15 @@ int main (int argc, char **argv)
 #ifdef REAL_UNIX_SYSTEM
    if (SLang_Version != SLANG_VERSION)
      {
-	fprintf (stderr, "\n\n******\n\
-slang.h does not match slang library version.  Did you install slang as\n\
-as a shared library?  Did you run ldconfig?  You have an installation problem\n\
-and you will need to check the SLANG variables in the Makefile and properly\n\
-set them.  Also try: make clean; make\n******\n\n");
+	fprintf (stderr, "\n\n******\n");
+	fprintf (stderr, "\
+slang.h (version=%ld) does not match the slang library version (%ld)\n\
+Did you install slang as a shared library?  Did you run ldconfig?\n\
+Perhaps you need to set the RPATH variable in the Makefile.\n\
+You have an installation problem and you will need to check the SLANG\n\
+variables in the Makefile and properly set them.\n\
+Also try: make clean; make\n", (long)SLANG_VERSION, (long)SLang_Version);
+	fprintf (stderr, "******\n\n");
 	return FAILURE;
      }
 #endif
