@@ -1,5 +1,5 @@
 /* -*- mode: c; mode: fold -*- */
-/* Copyright (c) 1992, 1998 John E. Davis
+/* Copyright (c) 1992, 1998, 2000, 2002, 2003, 2004, 2005, 2006 John E. Davis
  * This file is part of JED editor library source.
  *
  * You may distribute this file under the terms the GNU General Public
@@ -38,9 +38,11 @@
 #include "search.h"
 #include "hooks.h"
 #include "abbrev.h"
+#include "indent.h"
 #if JED_HAS_SUBPROCESSES
 # include "jprocess.h"
 #endif
+#include "sig.h"
 
 #if JED_HAS_LINE_ATTRIBUTES
 # include "lineattr.h"
@@ -49,7 +51,6 @@
 /*}}}*/
 
 /* fast insert/delete not implemented */
-#define USE_FAST_INSDEL 0
 
 /*{{{ Global variables */
 
@@ -58,6 +59,7 @@ int Blink_Flag = 1;
 int Indented_Text_Mode = 0;	       /* if non zero, intent line after wrap */
 int Kill_Line_Feature = 1;	       /* non-zero means kill through eol if bolp  */
 int Jed_Tab_Default = 8;
+int Jed_Case_Search_Default = 0;
 int Jed_Wrap_Column = 72;
 int Jed_Suspension_Not_Allowed = 0;
 int Jed_Use_Tabs = 1;
@@ -72,72 +74,6 @@ static char *End_Of_Buffer_Error = "End Of Buffer.";
 /*}}}*/
 /*{{{ Static functions */
 
-#if USE_FAST_INSDEL
-
-static void fast_ins(char ch) /*{{{*/
-{
-}
-
-/*}}}*/
-
-static void fast_del() /*{{{*/
-{
-}
-
-/*}}}*/
-
-/* return 1 if line extends beyond the screen */
-static int long_line(void)
-{
-   int p, col;
-   
-   col = Screen_Col;
-   p = Point;
-   Point = CLine->len - 1;
-   if ((*(CLine->data + Point) != '\n') || (CBuf == MiniBuffer)) Point++;
-   if (calculate_column() >= JWindow->width - 1)
-     {
-	Point = p;
-	Screen_Col = col;
-	return(1);
-     }
-   Screen_Col = col;
-   Point = p;
-   return(0);
-}
-
-/* check from point to end of line looking for tabs */
-static int tabs_present(void)
-{
-   unsigned char *p, *pmax;
-   
-   if (!Buffer_Local.tab) return(0);
-   pmax = CLine->data + CLine->len;
-   p = CLine->data + Point;
-   
-   while (p < pmax) if (*p++ == '\t') return(1);
-   return(0);
-}
-
-static int fast_not_ok (unsigned char ch)
-{
-   return 1;
-
-   return ((Repeat_Factor != NULL) || JWindow->trashed
-	   || (ch < 32) || (ch > 126) || Input_Buffer_Len
-	   || Suspend_Screen_Update
-	   || (Screen_Col >= JWindow->width - 2) || (JWindow->column > 1)
-	   || Batch || (CLine->len == 0)
-	   || ((CBuf->marks != NULL) && Wants_Attributes)
-#if JED_HAS_LINE_ATTRIBUTES
-	   || ((CLine->next != NULL) && (CLine->next->flags & JED_LINE_HIDDEN))
-#endif	   
-	   || input_pending(&Number_Zero) || Executing_Keyboard_Macro
-	   || tabs_present()
-	   || (Read_This_Character != NULL));
-}
-#endif				       /* USE_FAST_INSDEL */
-
 #if JED_HAS_LINE_ATTRIBUTES
 static int check_line_attr_no_modify (Line *l)
 {
@@ -177,7 +113,7 @@ static int next_visible_lines (int n)
    if (i) Point = 0;
    return i;
 #else
-   return nextline (&n);
+   return jed_down(n);
 #endif
 }
 
@@ -208,49 +144,49 @@ static int prev_visible_lines (int n)
    if (i) eol ();			       /* leave point at eol */
    return i;
 #else
-   return prevline (&n);
+   return jed_up (n);
 #endif
 }
 
 static int prev_visible_chars (int n)
 {
 #if JED_HAS_LINE_ATTRIBUTES
-   push_mark ();
-   n = backwchars (&n);
+   jed_push_mark ();
+   n = jed_left (n);
    if (n && (CLine->flags & JED_LINE_HIDDEN))
      {
 	jed_skip_hidden_lines_backward (&Number_One);
 	if (CLine->flags & JED_LINE_HIDDEN)
 	  {
-	     pop_mark (&Number_One);
+	     jed_pop_mark (1);
 	     return 0;
 	  }
      }
-   pop_mark (&Number_Zero);
+   jed_pop_mark (0);
    return n;
 #else
-   return backwchars (&n);
+   return jed_left (n);
 #endif
 }
 
 static int next_visible_chars (int n)
 {
 #if JED_HAS_LINE_ATTRIBUTES
-   push_mark ();
-   n = forwchars (&n);
+   jed_push_mark ();
+   n = jed_right (n);
    if (n && (CLine->flags & JED_LINE_HIDDEN))
      {
 	jed_skip_hidden_lines_forward (&Number_One);
 	if (CLine->flags & JED_LINE_HIDDEN)
 	  {
-	     pop_mark (&Number_One);
+	     jed_pop_mark (1);
 	     return 0;
 	  }
      }
-   pop_mark (&Number_Zero);
+   jed_pop_mark (0);
    return n;
 #else
-   return forwchars (&n);
+   return jed_right (n);
 #endif
 }
 
@@ -288,7 +224,7 @@ static void next_line_prev_line_helper (int *gcp, int *ret, int dr, FVOID_STAR f
    *ret = 1;
 #if 0
    *ret = (JWindow->trashed 
-	   || (CLine == JScreen[JWindow->top + dr - 1].line)
+	   || (CLine == JScreen[JWindow->sy + dr].line)
 #if JED_HAS_LINE_ATTRIBUTES
 	   || (CLine->flags & JED_LINE_IS_READONLY)
 #endif
@@ -300,10 +236,11 @@ static void eob_bob_error (int f)
 {   
    char *str;
    
-   if (CBuf->bob_eob_error_hook != NULL)
+   if ((CBuf->buffer_hooks != NULL)
+       && (CBuf->buffer_hooks->bob_eob_error_hook != NULL))
      {
 	SLang_push_integer (f);
-	SLexecute_function (CBuf->bob_eob_error_hook);
+	SLexecute_function (CBuf->buffer_hooks->bob_eob_error_hook);
 	return;
      }
    
@@ -323,11 +260,6 @@ static void eob_bob_error (int f)
 
 int delete_char_cmd (void)
 {
-   int upd, n;
-#if USE_FAST_INSDEL
-   char ch;
-#endif
-
    CHECK_READ_ONLY
 #if 0
      ;
@@ -344,45 +276,13 @@ int delete_char_cmd (void)
 	return(0);
      }
    
-#if USE_FAST_INSDEL
-   ch = *(CLine->data + Point);
-   if (fast_not_ok ((unsigned char) ch) || long_line () || *tt_Term_Cannot_Insert)
-     {
-#endif
-	if ((Point == 0) && eolp()
-#if JED_HAS_LINE_ATTRIBUTES
-	    && ((CLine->prev != NULL)
-		&& (0 == (CLine->prev->flags & JED_LINE_IS_READONLY)))
-#endif
-	    ) n = 1;
-	else n = 0;
-	/* same effect, update looks better */
-	n = backwchars(&n);
-	del();
-	forwchars(&n);
-	upd = 1;
-#if USE_FAST_INSDEL
-     }
-   else
-     {
-	upd = ((CBuf->flags & BUFFER_MODIFIED) == 0);
-	fast_del();
-	tt_delete_char ();
-     }
-#endif
-   return(upd);
+   (void) jed_del_wchar ();
+   return 1;
 }
 
 int backward_delete_char_cmd()
 {
-#if USE_FAST_INSDEL
-   char ch;
-   int ret;
-#endif
-   CHECK_READ_ONLY
-#if 0
-     ;
-#endif
+   /* CHECK_READ_ONLY */
 
    if (bobp())
      {
@@ -401,40 +301,21 @@ int backward_delete_char_cmd()
 	if (check_line_attr_no_modify (CLine->prev))
 	  return 0;
 #endif
-	backwchars(&Number_One);
-	return delete_char_cmd();
-     }
-   
-#if USE_FAST_INSDEL
-   ch = *(CLine->data + (Point - 1));
-   if (fast_not_ok ((unsigned char) ch) || long_line () || *tt_Term_Cannot_Insert)
-     {
-#endif
-	backwchars(&Number_One);
-	del();
-	return 1;
-#if USE_FAST_INSDEL
      }
 
-   ret = ((CBuf->flags & BUFFER_MODIFIED) == 0);
-   backwchars(&Number_One);
-   tt_putchar('\b');
-   Screen_Col--;
-   fast_del();
-   tt_delete_char();
-   return(ret);
-#endif
+   jed_left (1);
+   if (-1 == jed_del_wchar ())
+     jed_right (1);
+   return 1;
 }
 
 int backward_delete_char_untabify()
 {
    unsigned char *p;
    int n;
+
+   /* CHECK_READ_ONLY */
    
-   CHECK_READ_ONLY
-#if 0
-     ;
-#endif
 #if JED_HAS_LINE_ATTRIBUTES
    if (check_line_attr_no_modify (CLine))
      return 0;
@@ -445,10 +326,12 @@ int backward_delete_char_untabify()
      return backward_delete_char_cmd();
    
    n = calculate_column() - 1;
-   backwchars(&Number_One);
-   del();
+   jed_left (1);
+   if (-1 == jed_del_wchar ())
+     return -1;
+
    n = n - calculate_column();
-   ins_char_n_times(' ', n);
+   (void) jed_insert_wchar_n_times (' ', n);
    
    return(1);
 }
@@ -467,16 +350,17 @@ int newline_and_indent ()
 	   && eolp () && check_line_attr_no_modify (CLine->next)))
      return 0;
 #endif
-   
-   if (CBuf->newline_indent_hook != NULL)
+
+   if ((CBuf->buffer_hooks != NULL)
+       && (CBuf->buffer_hooks->newline_indent_hook != NULL))
      {
 	in_function = 1;
-	SLexecute_function(CBuf->newline_indent_hook);
+	SLexecute_function(CBuf->buffer_hooks->newline_indent_hook);
 	in_function = 0;
 	return 1;
      }
-   newline();
-   indent_line();
+   if (0 == jed_insert_newline ())
+     indent_line();
    return(1);
 }
 
@@ -492,7 +376,7 @@ int newline (void)
 
    if (CBuf == MiniBuffer) return exit_minibuffer();
    
-   insert_string ("\n");
+   (void) jed_insert_newline ();
    return(1);
 }
 
@@ -509,15 +393,25 @@ int newline_cmd (void)
 
 /*}}}*/
 
+static int execute_is_ok_hook (SLang_Name_Type *hook)
+{
+   int ret;
+   if (-1 == SLexecute_function (hook))
+     return -1;
+   if (-1 == SLang_pop_integer (&ret))
+     return -1;
+   return (ret != 0);
+}
+
 /*{{{ ins_char_cmd */
 int ins_char_cmd (void)
 {
-   char ch;
+   unsigned char ch;
    int wrap = Jed_Wrap_Column;
    int do_blink;
-#if USE_FAST_INSDEL
-   int ret, ll;
-#endif
+   int did_abbrev = 0;
+   SLang_Name_Type *wrapok_hook;
+
    CHECK_READ_ONLY
 #if 0
      ;
@@ -527,7 +421,7 @@ int ins_char_cmd (void)
      return 0;
 #endif
    
-   ch = (char) SLang_Last_Key_Char;
+   ch = SLang_Last_Key_Char;
    
    if (ch == '\n')
      {
@@ -538,19 +432,44 @@ int ins_char_cmd (void)
 #if JED_HAS_ABBREVS
    if (CBuf->flags & ABBREV_MODE)
      {
-	expand_abbrev (ch);
+	if (-1 == (did_abbrev = jed_expand_abbrev (ch)))
+	  return -1;
      }
 #endif
    
-   if ((CBuf->flags & OVERWRITE_MODE) && !eolp()) del();
+   if ((CBuf->flags & OVERWRITE_MODE) && !eolp())
+     {
+	/* FIXME: jed_del_wchar should be called for the last byte of a
+	 * UTF-8 sequence
+	 */
+	if ((did_abbrev == 0)
+	    && (-1 == jed_del_wchar ()))
+	  return -1;
+     }
+
+   /* It is ok to use Point as an estimator of the current column.  This 
+    * avoids the more expensive call to calculate_column.
+    */
+   if (CBuf->buffer_hooks != NULL) 
+     wrapok_hook = CBuf->buffer_hooks->wrapok_hook;
+   else
+     wrapok_hook = NULL;
    
-   if (((ch == ' ') || (Point >= wrap)) && (CBuf->modes & WRAP_MODE)
-       && (calculate_column() > wrap)) /* JWindow->width */
+   if (((ch == ' ') || (Point >= wrap)) 
+       && ((CBuf->modes & WRAP_MODE) || (wrapok_hook != NULL))
+       && (calculate_column() > wrap)
+       && ((wrapok_hook == NULL)
+	   || (1 == execute_is_ok_hook (wrapok_hook))))
      {
 	unsigned int this_line_num = LineNum;
-	ins(ch);
-	wrap_line(0);			/* do not format--- just wrap */
 	
+	if ((did_abbrev == 0)
+	    && (-1 == jed_insert_byte (ch)))
+	  return -1;
+
+	if (-1 == wrap_line(0))	       /* do not format--- just wrap */
+	  return -1;		       /* line isn't wrapable */
+
 	/* There is a bug involving wrapping a very long line containing
 	 * no whitespace and then we try to insert a character.  This work
 	 * arounds the bug.
@@ -559,15 +478,15 @@ int ins_char_cmd (void)
 	    && (ch == ' ')
 	    && (calculate_column () > wrap))
 	  {
-	     int one = 1;
-	     if (0 == forwchars (&one))
+	     if (0 == jed_right (1))
 	       newline ();
 	  }
 
-	if (Indented_Text_Mode) indent_line ();
-	if (CBuf->wrap_hook != NULL)
-	  SLexecute_function(CBuf->wrap_hook);
-	
+	if ((CBuf->buffer_hooks != NULL)
+	    && (CBuf->buffer_hooks->wrap_hook != NULL))
+	  SLexecute_function(CBuf->buffer_hooks->wrap_hook);
+	else if (Indented_Text_Mode) indent_line ();
+
 	return(1);
      }
    
@@ -575,50 +494,18 @@ int ins_char_cmd (void)
 		 && ((CBuf->syntax_table->char_syntax[(unsigned char) ch] & CLOSE_DELIM_SYNTAX)))
 		|| ((ch == ')') || (ch == '}') || (ch == ']')))
 	       && !input_pending(&Number_Zero));
-#if USE_FAST_INSDEL
-   if ((ch == ' ') || fast_not_ok ((unsigned char) ch) || (CBuf->flags & OVERWRITE_MODE)
-       || (*tt_Term_Cannot_Insert && !eolp()))
-     {
-#endif
-	ins(ch);
-	if (do_blink) blink_match ();
-	return 1;
-#if USE_FAST_INSDEL
-     }
-   
-   
-   ret = ((CBuf->flags & BUFFER_MODIFIED) == 0) || (User_Prefers_Line_Numbers > 1);
-   
-   if (long_line ()) ll = 1; else ll = 0;
-   ret = ret || ll;
-   
-   /* fast_ins goes first so that screen structure is updated now
-    for following calls to use. */
-   
-   fast_ins(ch);
-   if (!eolp() || ll)
-     {
-	tt_begin_insert();
-	tt_putchar(ch);
-	tt_end_insert();
-	if (ll) register_change(0);
-     }
-   else tt_putchar(ch);
-   Screen_Col++;
-   
+   if (did_abbrev == 0)
+     (void) jed_insert_byte (ch);
    if (do_blink) blink_match ();
-
-   return(ret);
-#endif
+   return 1;
 }
 
 /*}}}*/
 /*{{{ quoted_insert */
 int quoted_insert()
 {
-   char ch;
-   int lerr = SLang_Error;
-   
+   SLwchar_Type ch;
+
    CHECK_READ_ONLY
      if (*Error_Buffer || SLKeyBoard_Quit) return(0);
    if (Repeat_Factor != NULL)
@@ -633,18 +520,19 @@ int quoted_insert()
 	SLang_Key_TimeOut_Flag = 0;
      }
    
-   
-   SLang_Error = lerr;  /* ^G may set it */
+   if (SLang_get_error () == SL_USER_BREAK)
+     SLang_set_error (0);
    
    if ((ch == '\n') && (CBuf == MiniBuffer))
      {
-	ins('\n');
+	(void) _jed_ins_byte ('\n');
 	/* msg_error("Not allowed!"); */
 	return (1);
      }
    
    SLKeyBoard_Quit = 0;
-   ins_char_n_times(ch, 1);
+   if (-1 == jed_insert_wchar_n_times(ch, 1))
+     return -1;
    
    
    if ((CBuf->syntax_table != NULL)
@@ -657,6 +545,8 @@ int quoted_insert()
 /*}}}*/
 
 /*{{{ kill_line */
+
+/* FIXME: MULTIBYTE unsafe */
 int kill_line (void)
 {
    int n, pnt, flag = 0;
@@ -669,7 +559,7 @@ int kill_line (void)
 	return(0);
      }
    
-   push_mark();
+   jed_push_mark();
    push_spot();
    pnt = Point;
    eol();
@@ -685,9 +575,9 @@ int kill_line (void)
 	if (!pnt && (CLine->prev != NULL) && (CLine->next != NULL))
 	  {
 	     flag = 1;
-	     forwchars(&Number_One);
+	     jed_right (1);
 	  }
-	else n += forwchars(&Number_One);
+	else n += jed_right (1);
 	
      }
    
@@ -697,9 +587,9 @@ int kill_line (void)
      }
    else copy_to_pastebuffer();
    pop_spot();
-   if (flag) n += backwchars(&Number_One);
-   generic_deln(&n);
-   if (flag) forwchars(&Number_One);
+   if (flag) n += jed_left (1);
+   (void) jed_generic_del_nbytes (n);
+   if (flag) (void) jed_right (1);
    return(1);
 }
 
@@ -747,13 +637,11 @@ int next_line_cmd (void)
 
 int previous_char_cmd (void)
 {
-   int pnt;
-   
+   int b;
    Cursor_Motion = 1;
-   pnt = Point - 1;
    
    /* check_last_key_function ((FVOID_STAR) previous_char_cmd); */
-   
+   b = bolp ();
    if (1 != prev_visible_chars (1))
      {
 	eob_bob_error (-2);
@@ -761,7 +649,7 @@ int previous_char_cmd (void)
      }
    
    Goal_Column = calculate_column();
-   return (pnt == -1) || JWindow->trashed;
+   return b || JWindow->trashed;
 }
 
 int next_char_cmd ()
@@ -775,9 +663,9 @@ int next_char_cmd ()
 	eob_bob_error (2);
 	return 1;
      }
-   
+
    Goal_Column = calculate_column ();
-   return !Point || JWindow->trashed;  /* Point = 0 ==> moved a line */
+   return JWindow->trashed || bolp ();  /* Point = 0 ==> moved a line */
 }
 
 /*}}}*/
@@ -791,7 +679,7 @@ int eol_cmd (void)
        && (0 == (CLine->flags & JED_LINE_IS_READONLY))
 #endif
        )
-     trim_whitespace();
+     jed_trim_whitespace();
    return(1);
 }
 
@@ -878,6 +766,8 @@ int pagedown_cmd()
 	update (NULL, 0, 0, 1);
 	if (JWindow->trashed) return next_visible_lines (n);
      }
+
+   /* This is ugly. */
    
    this_line = LineNum;
    this_point = Point;
@@ -890,12 +780,15 @@ int pagedown_cmd()
    
    goto_column1 (&col);
    
-   if ((Last_Key_Function == (FVOID_STAR) pageup_cmd) && is_line_visible (Last_Page_Line))
+   if ((Last_Key_Function == (FVOID_STAR) pageup_cmd)
+       && (Jed_This_Key_Function == (FVOID_STAR) pagedown_cmd))
      {
 	goto_line (&Last_Page_Line);
-	Point = Last_Page_Point;
+	if (Last_Page_Point < CLine->len)
+	  Point = Last_Page_Point;
      }
-   else if (CLine->next == NULL) eol(); else Point = 0;
+   else if (CLine->next == NULL) eol(); 
+   else bol ();
    
    Last_Page_Line = this_line;
    Last_Page_Point = this_point;
@@ -941,40 +834,47 @@ int pageup_cmd (void)
    (void) goto_column1(&col);
    recenter(&JWindow->rows);
    
-   if ((Last_Key_Function == (FVOID_STAR) pagedown_cmd) && is_line_visible (Last_Page_Line))
+   if ((Last_Key_Function == (FVOID_STAR) pagedown_cmd) 
+       && (Jed_This_Key_Function == (FVOID_STAR) pageup_cmd))
      {
 	goto_line (&Last_Page_Line);
-	Point = Last_Page_Point;
+	if (Last_Page_Point < CLine->len)
+	  Point = Last_Page_Point;
      }
-   else Point = 0; /* something like: Point = point_column(JWindow->column) better? */
+   else 
+     bol (); /* something like: Point = point_column(JWindow->column) better? */
    Last_Page_Line = this_line;
    Last_Page_Point = this_point;
    return(1);
 }
 
-
-int scroll_right()
-{
-   if (JWindow->column == 1) return(0);
-   
-   if ((JWindow->column = JWindow->column - JWindow->width / 2) < 1)
-     {
-	JWindow->column = 1;
-     }
-   
-   touch_window();
-   return(1);
-}
-
-int scroll_left ()
-{
-   JWindow->column = JWindow->column + JWindow->width / 2;
-   touch_window();
-   return(1);
-}
-
 /*}}}*/
 
+int jed_scroll_left_cmd (void)
+{
+   int dc = JWindow->width/2;
+   if (Repeat_Factor != NULL)
+     {
+	dc = *Repeat_Factor;
+	Repeat_Factor = NULL;
+     }
+   jed_scroll_left (dc);
+   return 1;
+}
+
+int jed_scroll_right_cmd (void)
+{
+   int dc = JWindow->width/2;
+   if (Repeat_Factor != NULL)
+     {
+	dc = *Repeat_Factor;
+	Repeat_Factor = NULL;
+     }
+   jed_scroll_right (dc);
+   return 1;
+}
+
+
 /*}}}*/
 
 /*{{{ Whitespace and Indention functions */
@@ -994,9 +894,9 @@ void insert_whitespace(int *n)
 	i = c1 / tab;
 	k = c2 / tab - i;
 	if (k) nspace = c2 - (i + k) * tab;
-	ins_char_n_times('\t', k);
+	(void) jed_insert_wchar_n_times('\t', k);
      }
-   ins_char_n_times(' ', nspace);
+   (void) jed_insert_wchar_n_times(' ', nspace);
 }
 
 /* get indent value of current line, n is the column */
@@ -1020,49 +920,24 @@ unsigned char *get_current_indent(int *np)
    return p;
 }
 
-
-int trim_whitespace ()
+#define IS_WHITESPACE(p) ((*(p)==' ')||(*(p)=='\t'))
+int jed_trim_whitespace ()
 {
-   register unsigned char *p, *pmax, ch;
-   int n, save_point, len;
-   
-   CHECK_READ_ONLY
-     len = CLine->len;
-   if (len == 0) return(0);
-   
-   save_point = Point;
-   if (Point == len) Point--;
-   
-   /* Note that we save the point before --ing it.  This way the comparison
-    made below will effectively restore it if pointing at non whitespace */
-   
+   int n;
+   unsigned char *p;
+
+   /* CHECK_READ_ONLY */
+
    p = CLine->data + Point;
-   if (Point && (*p == '\n') && (CBuf != MiniBuffer))
-     {
-	Point--;
-	p--;
-     }
+   if ((0 == eolp ()) && (0 == IS_WHITESPACE(p)))
+     return 0;
+
+   (void) jed_skip_whitespace ();
+   n = Point;
    
-   while ((Point > 0) && (((ch = *p) == '\t') || (ch == ' ')))
-     {
-	Point--;
-	p--;
-     }
-   
-   if (save_point != Point) if (!eolp() && ((*p != ' ') && (*p != '\t')))
-     {
-	Point++;
-	p++;
-     }
-   
-   n = 0;
-   /* this needs to be inlined.  Actually for undo, I need del(n)! */
-   pmax = CLine->data + CLine->len;
-   while ((p < pmax) && ((ch = *p) != '\n')
-	  && ((ch == ' ') || (ch == '\t'))) p++, n++;
-   
-   deln(&n);
-   return(1);
+   (void) jed_bskip_whitespace ();
+   jed_del_nbytes (n - Point);
+   return 1;
 }
 
 /* indent line to column n */
@@ -1074,8 +949,8 @@ void indent_to(int n)
    
    if (n != m)
      {
-	Point = 0;
-	trim_whitespace();
+	bol ();
+	jed_trim_whitespace();
 	if (n >= 0) insert_whitespace(&n);
      }
 }
@@ -1087,15 +962,16 @@ int indent_line ()
    
    if (in_function) return -1;
    
-   if (CBuf->indent_hook != NULL)
+   if ((CBuf->buffer_hooks != NULL)
+       && (CBuf->buffer_hooks->indent_hook != NULL))
      {
 	in_function = 1;
-	SLexecute_function(CBuf->indent_hook);
+	SLexecute_function(CBuf->buffer_hooks->indent_hook);
 	in_function = 0;
 	return 1;
      }
    
-   CHECK_READ_ONLY
+   /* CHECK_READ_ONLY */
      
      if (CLine == CBuf->beg) return(0);
    
@@ -1147,152 +1023,78 @@ void goto_column(int *c)
 /*{{{ skip_whitespace */
 
 /* does not leave current line */
-unsigned int skip_whitespace()
+unsigned char *jed_skip_whitespace (void)
 {
    unsigned char *p, *pmax;
-   
-   if (CLine->len == 0) return('\n');
-   
+
    p = CLine->data + Point;
-   eol();
-   pmax = CLine->data + Point;
-   while(p < pmax)
-     {
-	if ((*p != ' ') && (*p != '\t')) break;
-	p++;
-     }
-   Point = (int) (p - CLine->data);
-   if (Point == CLine->len) return '\n';
-   return(*p);
+   pmax = jed_eol_position (CLine);
+
+   while ((p < pmax) && IS_WHITESPACE(p))
+     p = jed_multibyte_chars_forward (p, pmax, 1, NULL, 0);
+   
+   jed_position_point (p);
+   return p;
 }
 
-/*}}}*/
-/*{{{ skip_chars1 */
-void skip_chars1(char *range, int reverse)
+unsigned char *jed_bskip_whitespace (void)
 {
-   unsigned char lut[256];
-   unsigned char *p, *pmax;
-   Line *line;
-   int n = 0;
-   
-   SLmake_lut(lut, (unsigned char *) range, (unsigned char) reverse);
-   /* more flexibility is achieved with following line */
-   if (reverse && lut['\n']) lut['\n'] = 0;
-   
-   line = CLine;
-   while (line != NULL)
-     {
-	CLine = line;
-	p = line->data + Point;
-	pmax = line->data + line->len;
-	
-	while (p < pmax)
-	  {
-	     if (0 == lut[*p])
-	       {
-		  Point = (int)(p - CLine->data);
-                  LineNum += n;
-		  return;
-	       }
-	     p++;
-	  }
-	Point = 0;
-	line = CLine->next;
-        n++;
-     }
-   eob();
-}
+   unsigned char *p, *pmin, *pmax;
 
-/*}}}*/
-/*{{{ bskip_chars1 */
+   pmin = CLine->data;
+   pmax = p = CLine->data + Point;
 
-void bskip_chars1(char *range, int reverse)
-{
-   unsigned char lut[256];
-   unsigned char *p;
-   Line *line = CLine;
-   int n = 0;
-   
-   SLmake_lut(lut, (unsigned char *) range, (unsigned char) reverse);
-   /* more flexibility is achieved with following line */
-   if (reverse && lut['\n']) lut['\n'] = 0;
-   
-   while (1)
+   while (p > pmin) 
      {
-	if (Point == 0)
-	  {
-	     if (lut['\n'] == 0) return;
-	     if (NULL == (line = CLine->prev)) break;
-	     Point = line->len - 1;
-             n++;
-	  }
-	
-	CLine = line;
-	
-	p = line->data + (Point - 1);
-	
-	while (p >= line->data)
-	  {
-	     if (0 == lut[*p])
-	       {
-		  Point = 1 + (int)(p - line->data);
-                  LineNum -= n;
-		  return;
-	       }
-	     
-	     p--;
-	  }
-	Point = 0;
-     }
-   bob();
-}
-
-/*}}}*/
-/*{{{ skip_chars */
-void skip_chars (char *range)
-{
-   if (*range == '^') skip_chars1(range + 1, 1);
-   else
-     {
-	if (*range == '\\') range++;
-	skip_chars1(range, 0);
-     }
-}
-
-/*}}}*/
-/*{{{ bskip_chars */
-void bskip_chars (char *range)
-{
-   if (*range == '^') bskip_chars1(range + 1, 1);
-   else
-     {
-	if (*range == '\\') range++;
-	bskip_chars1(range, 0);
+	unsigned char *p1;
+	p1 = jed_multibyte_chars_backward (pmin, p, 1, NULL, 0);
+	if (0 == IS_WHITESPACE(p1))
+	  break;
+	p = p1;
      }
+   jed_position_point (p);
+   return p;
 }
 
 /*}}}*/
 
 /*{{{ looking_at */
 
-/* returns < 0 if what is smaller than current thing, 0 if matched, pos otherwise */
-int looking_at(char *what)
+/* MULTIBYTE OK */
+int jed_looking_at (char *what)
 {
-   register unsigned char *p, *pmax;
-   register unsigned char *w = (unsigned char *) what, ch;
+   unsigned char *p, *pmax;
+   unsigned char *w, *wmax;
    Line *l = CLine;
+   int cs = Buffer_Local.case_search;
    
+   w = (unsigned char *) what;
+   wmax = w + strlen (what);
+
    p = l->data + Point;
    while (1)
      {
 	pmax = l->data + l->len;
-	while (((ch = *w) != 0) && (p < pmax))
+	
+	if (cs)
 	  {
-#define upcase(ch) (Case_Sensitive ? ch : UPPER_CASE(ch))
-	     if ((upcase(ch) != upcase(*p))) return 0;
-	     p++; w++;
+	     while ((w < wmax) && (p < pmax))
+	       {
+		  if (*w++ != *p++)
+		    return 0;
+	       }
 	  }
-	if (ch == 0) return 1;
+	else 
+	  {
+	     /* Here we have to use multibyte routines */
+	     while ((w < wmax) && (p < pmax))
+	       {
+		  if (0 != jed_multibyte_charcasecmp (&w, wmax, &p, pmax))
+		    return 0;
+	       }
+	  }
+	if (w == wmax) return 1;
+
 	l = l->next;
 	if (l == NULL) return 0;
 	p = l->data;
@@ -1305,51 +1107,96 @@ int looking_at(char *what)
 
 /*{{{ sys_spawn_cmd */
 
-int sys_spawn_cmd (void)
+int jed_spawn_fg_process (int (*f)(VOID_STAR), VOID_STAR cd)
 {
+   int status;
+   int inited;
+
    if ((Jed_Secure_Mode)
        || (Jed_Suspension_Not_Allowed))
      {
 	msg_error ("Access to shell denied.");
 	return -1;
      }
+
+   /* FIXME: X_Suspend_Hook should not be here.  Currently, this hook is
+    * used only by GUI jed, where suspension makes no sense.  Of course in
+    * this case, spawning a foreground process also does not make sense.
+    */
+   if (Batch || (X_Suspend_Hook != NULL))
+     return (*f) (cd);
+
+   SLsig_block_signals ();
+   inited = Jed_Display_Initialized;
+   SLsig_unblock_signals ();
    
-   if (Batch)
+   jed_reset_display();
+#if !defined(IBMPC_SYSTEM) && !defined(VMS)
+   jed_reset_signals ();
+#endif
+   reset_tty();
+
+   status = (*f) (cd);
+
+   if (inited)
      {
-	sys_suspend ();
-	return 0;
-     }
+#if !defined(IBMPC_SYSTEM) && !defined(VMS)
+	init_signals();
+#endif
+	if (-1 == init_tty())
+	  {
+	     exit_error ("Unable to initialize terminal.", 0);
+	  }
    
+	flush_input ();
+	jed_init_display ();
+     }
+
+   check_buffers();
+   return status;
+}
+
+static int suspend_func (void *unused)
+{
+   (void) unused;
+   
+   if (X_Suspend_Hook != NULL)
+    (*X_Suspend_Hook) ();
+   else
+     sys_suspend();
+   
+   return 0;
+}
+   
+int sys_spawn_cmd (void)
+{
    if (X_Suspend_Hook != NULL)
      {
 	(*X_Suspend_Hook) ();
-	return (0);
+	return 0;
      }
-   
-   if (-1 == SLang_run_hooks("suspend_hook", 0))
-     return 0;
 
-   jed_reset_display();
-   reset_tty();
-   sys_suspend();
-   if (-1 == init_tty())
-     {
-	exit_error ("Unable to initialize terminal.", 0);
-     }
+   if (jed_va_run_hooks ("_jed_suspend_hooks", JED_HOOKS_RUN_UNTIL_0, 0) <= 0)
+     return 0;
    
-   flush_input ();
-   jed_init_display ();
-   (void) SLang_run_hooks("resume_hook", 0);
-   check_buffers();
-   return(1);
+   (void) jed_spawn_fg_process (suspend_func, NULL);
+
+   (void) jed_va_run_hooks ("_jed_resume_hooks", JED_HOOKS_RUN_ALL, 0);
+
+   return 0;
 }
+   
 
 /*}}}*/
-/*{{{ quit_jed */
-int quit_jed(void)
-{
-   Buffer *b = CBuf;
    
+/*{{{ jed_quit_jed */
+int jed_quit_jed(int status)
+{
+   Buffer *b;
+
+   (void) jed_va_run_hooks ("_jed_quit_hooks", JED_HOOKS_RUN_ALL, 0);
+
+   b = CBuf;
    /* Any buffer marked with AUTO_SAVE_JUST_SAVE flag should be saved
     * if it has not already been.  That is what the flag is for and this
     * code fragment carries this out.
@@ -1368,6 +1215,8 @@ int quit_jed(void)
  	if (b->subprocess)
  	  jed_kill_process (b->subprocess - 1);
 #endif
+
+	(void) jed_unlock_buffer_file (b);
 	
 	b = b->next;
      }
@@ -1385,7 +1234,7 @@ int quit_jed(void)
    SLmalloc_dump_statistics ();
 #endif
    /* SLstring_dump_stats (); */
-   exit(0);
+   exit (status);
    return(1);
 }
 
@@ -1421,11 +1270,16 @@ int save_some_buffers (void)
 	       {
 		  tmp = CBuf;
 		  switch_to_buffer(b);
-		  while (b->narrow != NULL) widen_buffer(b);
-		  err = write_file_with_backup(b->dir, b->file);
+
+		  /* It should not be necessary to do this here.  Lower
+		   * level routines will do it.
+		   */
+		  /* while (b->narrow != NULL) widen_buffer(b); */
+
+		  err = jed_save_buffer_cmd ();
 		  switch_to_buffer(tmp);
 		  if (err < 0) return -1;
-		  b->flags &= ~BUFFER_MODIFIED;
+		  /* b->flags &= ~BUFFER_MODIFIED; */
 		  /* b->flags |= AUTO_SAVE_BUFFER; */
 		  b->hits = 0;
 	       }
@@ -1441,30 +1295,30 @@ int save_some_buffers (void)
 
 /*}}}*/
 /*{{{ exit_jed */
-int exit_jed (void)
+int jed_exit_jed (int status)
 {
    static int in_exit_jed = 0;
    
    if (in_exit_jed == 0)
      {
 	in_exit_jed = 1;
-	if (-1 == SLang_run_hooks("exit_hook", 0))
+	if (jed_va_run_hooks ("_jed_exit_hooks", JED_HOOKS_RUN_UNTIL_0, 0) <= 0)
 	  {
 	     in_exit_jed = 0;
-	     return 1;
+	     return -1;
 	  }
      }
-   
+
    in_exit_jed = 0;
-   if (SLang_Error) return 0;
+   if (SLang_get_error ())
+     return -1;
 
 #if JED_HAS_SUBPROCESSES
-   if (Num_Subprocesses
-       && (1 != jed_get_y_n ("Subprocesses exist.  Exit anyway")))
+   if (1 != jed_processes_ok_to_exit ())
      return 1;
 #endif
    
-   if (save_some_buffers() > 0) quit_jed();
+   if (save_some_buffers() > 0) jed_quit_jed(status);
    return 1;
 }
 
@@ -1472,3 +1326,7 @@ int exit_jed (void)
 
 /*}}}*/
 
+int jed_exit_jed_cmd (void)
+{
+   return jed_exit_jed (0);
+}
