@@ -1,4 +1,4 @@
-/* Copyright (c) 1992, 1998 John E. Davis
+/* Copyright (c) 1992, 1998, 2000, 2005, 2006 John E. Davis
  * This file is part of JED editor library source.
  *
  * You may distribute this file under the terms the GNU General Public
@@ -24,7 +24,7 @@ extern int write_region(char *);
 extern int read_file_pointer(int);
 extern int insert_file_pointer(VFILE *);
 extern int append_to_file(char *);
-extern int write_file_with_backup(char *, char *);
+extern int write_file_with_backup(char *);
 extern void visit_file(char *, char *);
 extern void fixup_dir(char *);
 extern char *dir_file_merge(char *, char *);
@@ -35,11 +35,10 @@ extern char *file_type(char *);
 extern void check_buffer(Buffer *);
 extern void set_file_trans(int *);
 
-#ifdef __unix__
-extern char *expand_link(char *);
-#endif
+extern char *jed_expand_link(char *);
+
 #ifdef REAL_UNIX_SYSTEM
-extern int get_inode_info (char *, int *, int *);
+extern int jed_get_inode_info (char *, int *, int *);
 extern int Jed_Backup_By_Copying;
 #endif
 #ifndef VMS
@@ -48,4 +47,13 @@ extern int jed_copy_file (char *, char *);
 extern void jed_set_umask (int);
 extern void jed_set_buffer_ctime (Buffer *);
 
-
+extern int jed_unlock_buffer_files (void);
+extern int jed_unlock_buffer_file (Buffer *);
+extern int jed_lock_buffer_file (Buffer *);
+extern int jed_lock_file (char *);
+extern int jed_unlock_file (char *);
+extern int jed_dirfile_to_dir_file (char *dirfile, char **dirp, char **filep);
+extern char *jed_standardize_filename (char *file);
+extern char *jed_dir_file_merge (char *dir, char *file);
+extern char *jed_get_canonical_pathname (char *);
+extern int jed_buffer_file_is_readonly (Buffer *);
