@@ -1,5 +1,5 @@
 /* -*- mode: C; mode: fold; -*- */
-/* Copyright (c) 1992, 1998 John E. Davis
+/* Copyright (c) 1992, 1998, 2000, 2002, 2003, 2004, 2005, 2006 John E. Davis
  * This file is part of JED editor library source.
  *
  * You may distribute this file under the terms the GNU General Public
@@ -21,20 +21,32 @@
 #include "sysdep.h"
 #include "search.h"
 #include "paste.h"
+#include "misc.h"
+#include "indent.h"
 
 /*}}}*/
 
+/* Note: In this file, "UTF-8 ASCII safe" means that the routine will work
+ * properly as long as we are dealing with ASCII characters for quotes and
+ * comment strings.
+ *
+ * "UTF-8 WORD UNsafe" indicates that the routine is not safe when non-ASCII
+ * characters are given the word syntax.
+ * 
+ * ASCII means character codes 0-127.
+ */
 Syntax_Table_Type *Default_Syntax_Table;
 
 static Syntax_Table_Type *Syntax_Tables;
 
+/* UTF-8 ASCII safe */
 static unsigned char *find_comment_end (Syntax_Table_Type *table,
 					unsigned char *p, unsigned char *pmax)
 {
    unsigned int len;
    char *str;
    unsigned char ch;
-   unsigned char quote;
+   /* unsigned char quote; */
 
    str = table->comment_stop;
    len = table->comment_stop_len;
@@ -45,16 +57,10 @@ static unsigned char *find_comment_end (Syntax_Table_Type *table,
    pmax -= (len - 1);
 
    ch = *str;
-   quote = table->quote_char;
+   /* quote = table->quote_char; */
 
    while (p < pmax)
      {
-	if (*p == quote)
-	  {
-	     p += 2;
-	     continue;
-	  }
-
 	if ((*p == ch)
 	    && (0 == strncmp (str, (char *)p, len)))
 	  return p + len;
@@ -65,6 +71,7 @@ static unsigned char *find_comment_end (Syntax_Table_Type *table,
    return NULL;
 }
 
+/* UTF-8 ASCII safe */
 static unsigned char *find_string_end (Syntax_Table_Type *table,
 				       unsigned char *p, unsigned char *pmax,
 				       unsigned char ch)
@@ -77,32 +84,124 @@ static unsigned char *find_string_end (Syntax_Table_Type *table,
      {
 	if (*p == quote)
 	  {
-	     p += 2;
+	     p = jed_multibyte_chars_forward (p+1, pmax, 1, NULL, 0);
 	     continue;
 	  }
-
 	if (*p++ == ch)
 	  return p;
      }
-
+   
    return NULL;
 }
 
+/* UTF-8 ASCII safe */
 static int is_fortran_comment (Line *l, Syntax_Table_Type *table)
 {
    return l->len && table->fortran_comment_chars[l->data[0]];
 }
 
+/* UTF-8 WORD unsafe */
+int _jed_is_eol_comment_start (Syntax_Table_Type *table, Line *l,
+			       unsigned char *p, unsigned char *pmax, unsigned int *com)
+{
+   unsigned int i;
+   int wsc;
+
+   if (p >= pmax)
+     return 0;
+
+   /* Most of the time, num_eol_comments is going to be <= 1, so just 
+    * use table->whatever.
+    */
+   wsc = table->flags & EOL_COMMENT_NEEDS_WHITESPACE;
+
+   for (i = 0; i < table->num_eol_comments; i++)
+     {
+	if (*p != table->eol_comment_starts[i][0])
+	  continue;
+	
+	if (p + table->eol_comment_lens[i] > pmax)
+	  continue;
+
+	if (0 == (table->flags & SYNTAX_NOT_CASE_SENSITIVE))
+	  {
+	     if (0 != strncmp ((char *) p, table->eol_comment_starts[i],
+			       table->eol_comment_lens[i]))
+	       continue;
+	  }
+	else
+	  {
+	     if (0 != jed_case_strncmp ((char *) p, table->eol_comment_starts[i],
+					table->eol_comment_lens[i]))
+	       continue;
+	  }
+	
+	if ((table->char_syntax[*p] & WORD_SYNTAX) || wsc)
+	  {
+	     unsigned char *p1;
+	     unsigned char ch1;
+
+	     p1 = p + table->eol_comment_lens[i];
+	     if (p1 < pmax)
+	       {
+		  ch1 = *p1;
+		  if (wsc)
+		    {
+		       if ((ch1 != ' ') && (ch1 != '\n') && (ch1 != '\t'))
+			 {
+			    /* This piece of code permits: #### */
+			    if ((ch1 != *p)
+				|| (table->char_syntax[ch1] & WORD_SYNTAX)
+				|| (p1 != p + 1))
+			      continue;
+			 }
+		    }
+		  else if (table->char_syntax[ch1] & WORD_SYNTAX)
+		    continue;
+	       }
+	     
+	     if (p > l->data)
+	       {
+		  ch1 = *(p-1);
+
+		  if (wsc)
+		    {
+		       if ((ch1 != ' ') && (ch1 != '\t'))
+			 {
+			    /* This piece of code permits: #### */
+			    if ((ch1 != *p)
+				|| (table->char_syntax[ch1] & WORD_SYNTAX)
+				|| (p1 != p + 1))
+			      continue;
+			 }
+		    }
+		  else if (table->char_syntax[ch1] & WORD_SYNTAX)
+		    continue;
+	       }
+	  }
+
+	if (com != NULL) 
+	  *com = i;
+
+	return 1;
+     }
+   
+   return 0;
+}
+
 static int parse_to_point1 (Syntax_Table_Type *table,
 			    Line *l,
 			    unsigned char *pmax)
 {
    unsigned char ch;
-   int quote,  sc,  cc,  flags;
+   int quote,  cc,  flags;
    unsigned char *p;
    unsigned int lflags;
-   unsigned char com_start_char, eol_com_start_char;
-   char *comment_start, *eol_comment_start;
+   unsigned char com_start_char;
+   char *comment_start;
+   unsigned short *syntax;
+   unsigned char *sc;
+   unsigned char sgml_stop_char;
 
    flags = table->flags;
    p = l->data;
@@ -113,8 +212,10 @@ static int parse_to_point1 (Syntax_Table_Type *table,
 	  return JED_LINE_HAS_EOL_COMMENT;
      }
 
-   sc = table->string_char;
+   sc = table->string_chars;
    cc = table->char_char;
+   syntax = table->char_syntax;
+   sgml_stop_char = table->sgml_stop_char;
 
    lflags = l->flags & JED_LINE_SYNTAX_BITS;
 
@@ -125,10 +226,20 @@ static int parse_to_point1 (Syntax_Table_Type *table,
 	     if (NULL == (p = find_comment_end (table, p, pmax)))
 	       return JED_LINE_IN_COMMENT;
 	  }
-	else if (lflags & JED_LINE_IN_STRING)
+	else if (lflags & JED_LINE_IN_STRING0)
 	  {
-	     if (NULL == (p = find_string_end (table, p, pmax, sc)))
-	       return JED_LINE_IN_STRING;
+	     if (NULL == (p = find_string_end (table, p, pmax, sc[0])))
+	       return JED_LINE_IN_STRING0;
+	  }
+	else if (lflags & JED_LINE_IN_STRING1)
+	  {
+	     if (NULL == (p = find_string_end (table, p, pmax, sc[1])))
+	       return JED_LINE_IN_STRING1;
+	  }
+	else if (lflags & JED_LINE_IN_HTML)
+	  {
+	     if (NULL == (p = find_string_end (table, p, pmax, sgml_stop_char)))
+	       return JED_LINE_IN_HTML;
 	  }
      }
 
@@ -136,10 +247,8 @@ static int parse_to_point1 (Syntax_Table_Type *table,
    quote = table->quote_char;
 
    comment_start = table->comment_start;
-   eol_comment_start = table->eol_comment_start;
 
    com_start_char = (comment_start == NULL) ? 0 : *comment_start;
-   eol_com_start_char = (eol_comment_start == NULL) ? 0 : *eol_comment_start;
 
    while (p < pmax)
      {
@@ -151,15 +260,54 @@ static int parse_to_point1 (Syntax_Table_Type *table,
 	  }
 
 	/* Search for a string delimiter, char delimiter, or comment start. */
-	if ((ch == sc) || (ch == cc))
+#if 1
+	/* Check comment syntax first */
+	if (syntax[ch] & COMMENT_SYNTAX)
+	  {
+	     if ((ch == com_start_char)
+		 && (comment_start != NULL)
+		 && (p + table->comment_start_len <= pmax)
+		 && (0 == strncmp ((char *) p, comment_start,
+				   table->comment_start_len)))
+	       {
+		  p = find_comment_end (table, p + table->comment_start_len, pmax);
+		  if (p == NULL)
+		    return JED_LINE_IN_COMMENT;
+
+		  continue;
+	       }
+
+	     if (_jed_is_eol_comment_start (table, l, p, pmax, NULL))
+	       return JED_LINE_HAS_EOL_COMMENT;
+	  }
+#endif	
+	if (syntax[ch] & (STRING_SYNTAX))
 	  {
 	     p = find_string_end (table, p + 1, pmax, ch);
 	     if (p == NULL)
-	       return JED_LINE_IN_STRING;
-
+	       {
+		  if (sc[0] == ch)
+		    return JED_LINE_IN_STRING0;
+		  if (sc[1] == ch)
+		    return JED_LINE_IN_STRING1;
+		  return JED_LINE_IN_STRING0;
+	       }
 	     continue;
 	  }
 
+	if (syntax[ch] & HTML_START_SYNTAX)
+	  {
+	     p = find_string_end (table, p+1, pmax, table->sgml_stop_char);
+	     if (p == NULL)
+	       return JED_LINE_IN_HTML;
+	     continue;
+	  }
+#if 0	
+	if (0 == (syntax[ch] & COMMENT_SYNTAX))
+	  {
+	     p++;
+	     continue;
+	  }
 	if ((ch == com_start_char)
 	    && (comment_start != NULL)
 	    && (p + table->comment_start_len <= pmax)
@@ -173,12 +321,9 @@ static int parse_to_point1 (Syntax_Table_Type *table,
 	     continue;
 	  }
 
-	if ((ch == eol_com_start_char)
-	    && (eol_comment_start != NULL)
-	    && (p + table->eol_comment_len <= pmax)
-	    && (0 == strncmp ((char *) p, eol_comment_start,
-			      table->eol_comment_len)))
+	if (_jed_is_eol_comment_start (table, l, p, pmax, NULL))
 	  return JED_LINE_HAS_EOL_COMMENT;
+#endif
 
 	p++;
      }
@@ -190,8 +335,7 @@ static void goto_effective_eol (Syntax_Table_Type *table) /*{{{*/
 {
    unsigned int flags;
    unsigned char *p, *pmax;
-   char *comment_start;
-   unsigned char ch;
+   unsigned short *syntax;
 
    flags = CLine->flags;
    if (0 == (flags & JED_LINE_HAS_EOL_COMMENT))
@@ -202,13 +346,13 @@ static void goto_effective_eol (Syntax_Table_Type *table) /*{{{*/
 
    if (table->flags & FORTRAN_TYPE)
      {
-	Point = 0;
+	bol ();
 
 	if (is_fortran_comment (CLine, table))
 	  return;
      }
 
-   if (NULL == (comment_start = table->eol_comment_start))
+   if (0 == table->num_eol_comments)
      {
 	/* ?? */
 	CLine->flags &= ~JED_LINE_HAS_EOL_COMMENT;
@@ -219,42 +363,28 @@ static void goto_effective_eol (Syntax_Table_Type *table) /*{{{*/
    /* We know that there is an EOL style comment somewhere on this line.  So,
     * just search for it.
     */
-   if (table->flags & COMMENT_IS_WORD)
-     {
-	/* FIXME: Handle this later .... */
-	eol ();
-	return;
-     }
-
    p = CLine->data;
    pmax = CLine->data + CLine->len;
-
-   ch = (unsigned char) *comment_start;
+   
+   syntax = table->char_syntax;
 
    while (p < pmax)
      {
-	if (*p != ch)
-	  {
-	     p++;
-	     continue;
-	  }
+	unsigned int i;
 
-	if (p + table->eol_comment_len > pmax)
-	  break;		       /* ??? */
-
-	if (0 != strncmp ((char *) p, comment_start, table->eol_comment_len))
+	if ((syntax[*p] & COMMENT_SYNTAX)
+	    && (_jed_is_eol_comment_start (table, CLine, p, pmax, &i)))
 	  {
-	     p++;
-	     continue;
-	  }
-
-	/* Is this really a comment?  Only one way to find out */
-	Point = (int) (p - CLine->data) + table->eol_comment_len;
-	if (JED_LINE_HAS_EOL_COMMENT == parse_to_point1 (table, CLine, 
-							 p + table->eol_comment_len))
-	  {
-	     Point = (int) (p - CLine->data);
-	     return;		       /* we are there */
+	     /* Is this really a comment?  Perhaps is occurs in a string
+	      * or in part of another word. Only one way to find out.
+	      */
+	     jed_position_point (p + table->eol_comment_lens[i]);
+	     if (JED_LINE_HAS_EOL_COMMENT == parse_to_point1 (table, CLine, 
+							      p + table->eol_comment_lens[i]))
+	       {
+		  jed_position_point (p);
+		  return;		       /* we are there */
+	       }
 	  }
 	p++;
      }
@@ -352,19 +482,18 @@ static int goto_comment_begin (Syntax_Table_Type *table)
 		 && (0 == strncmp (s, (char *)p, len))
 		 && (0 == parse_to_point1 (table, CLine, p)))
 	       {
-		  Point = (int) (p - pmin);
+		  jed_position_point (p);
 		  return 0;
 	       }
 
 	     p--;
 	  }
 
-	if (CLine->prev == NULL)
+	if (0 == jed_up (1))
 	  {
-	     Point = 0;
+	     bol ();
 	     return -1;
 	  }
-	CLine = CLine->prev; LineNum--; Point = CLine->len - 1;
      }
 }
 
@@ -392,19 +521,17 @@ static int goto_comment_end (Syntax_Table_Type *table)
 		 && (p + len <= pmax)
 		 && (0 == strncmp (s, (char *)p, len)))
 	       {
-		  p += len;
-		  Point = (int) (p - CLine->data);
+		  jed_position_point (p + len);
 		  return 0;
 	       }
 	     p++;
 	  }
-
-	if (CLine->next == NULL)
+	
+	if (0 == jed_down (1))
 	  {
 	     eol ();
 	     return -1;
 	  }
-	CLine = CLine->next; LineNum++; Point = 0;
      }
 }
 
@@ -432,18 +559,17 @@ static int goto_string_begin (Syntax_Table_Type *table,
 		      && (is_quoted (pmin, &p, quote)))
 		    continue;
 
-		  p++;
-		  Point = (int) (p - pmin);
+		  jed_position_point (p + 1);
 		  return 0;
 	       }
 	  }
-	if (CLine->prev == NULL)
+	
+	if ((table->flags & SINGLE_LINE_STRINGS)
+	    || (0 == jed_up (1)))
 	  {
-	     Point = 0;
+	     bol ();
 	     return -1;
 	  }
-
-	CLine = CLine->prev; LineNum--; Point = CLine->len - 1;
 	pmin = CLine->data;
 	p = pmin + Point;
      }
@@ -463,21 +589,19 @@ static int goto_string_end (Syntax_Table_Type *table, unsigned char ch)
 	  {
 	     if (*p == ch)
 	       {
-		  p++;		       /* go around it */
-		  Point = (int) (p - CLine->data);
+		  jed_position_point (p + 1);  /* go around it */
 		  return 0;
 	       }
 	     if (*p == quote) p++;
 	     p++;
 	  }
 
-	if (CLine->next == NULL)
+	if ((table->flags & SINGLE_LINE_STRINGS)
+	    || (0 == jed_down (1)))
 	  {
 	     eol ();
 	     return -1;
 	  }
-
-	CLine = CLine->next; LineNum++; Point = 0;
      }
 }
 
@@ -506,7 +630,7 @@ static int backward_goto_match (int count, unsigned char ch) /*{{{*/
    Syntax_Table_Type *table;
    unsigned int this_syntax;
    unsigned char *pmax;
-   unsigned char eol_com_char = 0, com_start_char = 0, com_stop_char = 0;
+   unsigned char com_start_char, com_stop_char;
 
    if (NULL == (table = setup_for_match (&p, &ch, &want_ch)))
      return 0;
@@ -525,17 +649,19 @@ static int backward_goto_match (int count, unsigned char ch) /*{{{*/
 	in_comment = 1;
 	break;
 
-      case JED_LINE_IN_STRING:
-	in_string = 1;
+      case JED_LINE_IN_STRING0:
+	in_string = (int) table->string_chars[0];
+	break;
+      case JED_LINE_IN_STRING1:
+	in_string = (int) table->string_chars[1];
 	break;
      }
    Point--;
 
-   if (table->eol_comment_start != NULL)
-     eol_com_char = (unsigned char) table->eol_comment_start[0];
-
    if (table->comment_start != NULL)
      com_start_char = (unsigned char) table->comment_start[0];
+   else
+     com_start_char = 0;
 
    if (table->comment_stop != NULL)
      com_stop_char = (unsigned char) table->comment_stop[0];
@@ -578,20 +704,18 @@ static int backward_goto_match (int count, unsigned char ch) /*{{{*/
 		  if (in_comment) continue;
 		  if (in_string)
 		    {
-		       if (ch == table->string_char)
+		       if (in_string == (int) ch)
 			 {
 			    /* Match not found.  Leave point at beginning
 			     * of string and return -1.
 			     */
-			    Point = 1 + (int) (p - pmin);
+			    jed_position_point (p + 1);
 			    return -1;
 			 }
 		       continue;
 		    }
 
-		  Point = (int) (p - pmin);
-		  if (Point < 0)
-		    Point = 0;
+		  jed_position_point (p);
 
 		  if (-1 == goto_string_begin (table, ch, p))
 		    {
@@ -614,7 +738,7 @@ static int backward_goto_match (int count, unsigned char ch) /*{{{*/
 	       {
 		  if (level == 1)
 		    {
-		       Point = (int) (p - pmin) + 1;
+		       jed_position_point (p+1);
 		       if (ch == want_ch) return 1;
 		       return 0;
 		    }
@@ -643,12 +767,9 @@ static int backward_goto_match (int count, unsigned char ch) /*{{{*/
 		   * you lose.
 		   */
 		  p++;
-		  if ((ch == eol_com_char)
-		      && (table->eol_comment_start != NULL)
-		      && (p + table->eol_comment_len <= pmax)
-		      && (0 == strncmp ((char *) p, table->eol_comment_start, table->eol_comment_len)))
+		  if (_jed_is_eol_comment_start (table, CLine, p, pmax, NULL))
 		    {
-		       Point = (int) (p - pmin);
+		       jed_position_point (p);
 		       return -2;
 		    }
 
@@ -657,7 +778,7 @@ static int backward_goto_match (int count, unsigned char ch) /*{{{*/
 		      && (p + table->comment_start_len <= pmax)
 		      && (0 == strncmp ((char *) p, table->comment_start, table->comment_start_len)))
 		    {
-		       Point = (int) (p - pmin);
+		       jed_position_point (p);
 		       return -2;
 		    }
 		  p--;
@@ -681,7 +802,7 @@ static int backward_goto_match (int count, unsigned char ch) /*{{{*/
 		  continue;
 	       }
 
-	     Point = (int) (p - pmin);
+	     jed_position_point (p);
 	     if (-1 == goto_comment_begin (table))
 	       return 0;
 
@@ -690,15 +811,14 @@ static int backward_goto_match (int count, unsigned char ch) /*{{{*/
 	     p = pmin + (Point - 1);
 	  }
 
-	/* Move to the previous line. */
-	if (CLine->prev == NULL)
+	if (0 == jed_up (1))
 	  {
-	     Point = 0;
+	     bol ();
 	     break;
 	  }
-	CLine = CLine->prev; LineNum--; Point = CLine->len - 1;
 	count--;
-
+	if (table->flags & SINGLE_LINE_STRINGS)
+	  in_string = 0;
 	goto_effective_eol (table);
 	pmax = CLine->data + Point;
      }
@@ -711,7 +831,7 @@ static int backward_goto_match (int count, unsigned char ch) /*{{{*/
 	/* In this case, we went back as far as permitted.  Nothing much can be
 	 * said.
 	 */
-	Point = 0;
+	bol ();
 	return 2;
      }
 
@@ -731,7 +851,7 @@ static int forward_goto_match (unsigned char ch) /*{{{*/
    int in_string, in_comment, level;
    unsigned int this_syntax;
    Syntax_Table_Type *table;
-   unsigned char eol_com_char, com_start_char;
+   unsigned char com_start_char;
 
    if (NULL == (table = setup_for_match (&p, &ch, &want_ch)))
      return 0;
@@ -751,15 +871,15 @@ static int forward_goto_match (unsigned char ch) /*{{{*/
 	in_comment = JED_LINE_IN_COMMENT;
 	break;
 
-      case JED_LINE_IN_STRING:
-	in_string = 1;
+      case JED_LINE_IN_STRING0:
+	in_string = (int) table->string_chars[0];
+	break;
+      case JED_LINE_IN_STRING1:
+	in_string = (int) table->string_chars[1];
 	break;
      }
 
-   eol_com_char = com_start_char = 0;
-
-   if (table->eol_comment_start != NULL)
-     eol_com_char = (unsigned char) table->eol_comment_start[0];
+   com_start_char = 0;
 
    if (table->comment_start != NULL)
      com_start_char = (unsigned char) table->comment_start[0];
@@ -808,7 +928,7 @@ static int forward_goto_match (unsigned char ch) /*{{{*/
 			   && (p + table->comment_start_len <= pmax)
 			   && (0 == strncmp ((char *) p, table->comment_start, table->comment_start_len)))
 			 {
-			    Point = (int) (p - CLine->data);
+			    jed_position_point (p);
 			    if (-1 == goto_comment_end (table))
 			      return 0;
 
@@ -817,10 +937,7 @@ static int forward_goto_match (unsigned char ch) /*{{{*/
 			    continue;
 			 }
 		       
-		       if ((ch == eol_com_char)
-			   && (table->eol_comment_start != NULL)
-			   && (p + table->eol_comment_len <= pmax)
-			   && (0 == strncmp ((char *) p, table->eol_comment_start, table->eol_comment_len)))
+		       if (_jed_is_eol_comment_start (table, CLine, p, pmax, NULL))
 			 {
 			    p = pmax;
 			    continue;
@@ -836,12 +953,12 @@ static int forward_goto_match (unsigned char ch) /*{{{*/
 		  if (in_comment) continue;
 		  if (in_string)
 		    {
-		       if (ch == table->string_char)
+		       if ((int) ch == in_string)
 			 in_string = 0;
 		    }
 		  else
 		    {
-		       Point = (int) (p - CLine->data);
+		       jed_position_point (p);
 		       if (-1 == goto_string_end (table, ch))
 			 return -1;
 
@@ -861,7 +978,7 @@ static int forward_goto_match (unsigned char ch) /*{{{*/
 	       {
 		  if (level == 1)
 		    {
-		       Point = (int) (p - CLine->data) - 1;
+		       jed_position_point (p-1);
 		       if (ch == want_ch) return 1;
 		       return 0;
 		    }
@@ -879,7 +996,7 @@ static int forward_goto_match (unsigned char ch) /*{{{*/
 	/* Move to the next line. */
 	while (1)
 	  {
-	     if (CLine->next == NULL)
+	     if (0 == jed_down (1))
 	       {
 		  eol ();
 		  if (in_string) return -1;
@@ -889,7 +1006,9 @@ static int forward_goto_match (unsigned char ch) /*{{{*/
 		  return 0;
 	       }
 
-	     CLine = CLine->next; LineNum++; Point = 0;
+	     if (table->flags & SINGLE_LINE_STRINGS)
+	       in_string = 0;
+
 	     if ((0 == (table->flags & FORTRAN_TYPE))
 		 || (0 == is_fortran_comment (CLine, table)))
 	       break;
@@ -915,7 +1034,7 @@ static int find_matching_delimiter_1 (unsigned char ch, int nlines)
      return backward_goto_match (nlines, ch);
 }
 
-int find_matching_delimiter (int *ch)
+static int find_matching_delimiter (int *ch)
 {
    return find_matching_delimiter_1 ((unsigned char) *ch, 5000);
 }
@@ -932,20 +1051,24 @@ int goto_match (void) /*{{{*/
 
 /*}}}*/
 
-int parse_to_point (void) /*{{{*/
+static int parse_to_point (void) /*{{{*/
 {
-   int state;
-
    Syntax_Table_Type *table = CBuf->syntax_table;
    if (table == NULL) return 0;
 #if JED_HAS_LINE_ATTRIBUTES
    jed_syntax_parse_buffer (0);
 #endif
-   state = parse_to_point1 (table, CLine, CLine->data + Point);
+   
+   switch (parse_to_point1 (table, CLine, CLine->data + Point))
+     {
+      case JED_LINE_IN_COMMENT:
+      case JED_LINE_HAS_EOL_COMMENT:
+	return -2;
+      case JED_LINE_IN_STRING0:
+      case JED_LINE_IN_STRING1:
+	return -1;
+     }
 
-   if (state == JED_LINE_IN_COMMENT) return -2;
-   if (state == JED_LINE_IN_STRING) return -1;
-   if (state == JED_LINE_HAS_EOL_COMMENT) return -2;
    return 0;
 }
 
@@ -989,17 +1112,16 @@ void blink_match (void) /*{{{*/
 	unsigned int len;
 
 	matchp = Point;
-	Point = 0;
+	bol ();
 	strcpy(buf, "Matches ");
-	skip_whitespace();
-	if ((matchp == Point) && prevline(&Number_One))
+	(void) jed_skip_whitespace();
+	if ((matchp == Point) && jed_up(1))
 	  {
-	     Point = 0;
+	     bol ();
 	     safe_strcat (buf, 
 			  make_line_string(strbuf, sizeof(strbuf)), 
 			  sizeof (buf));
-	     nextline(&Number_One);
-	     Point = 0;
+	     jed_down(1);
 	  }
 
 	safe_strcat(buf, 
@@ -1022,12 +1144,12 @@ void blink_match (void) /*{{{*/
 
 /*}}}*/
 
-Syntax_Table_Type *find_syntax_table (char *name, int err) /*{{{*/
+Syntax_Table_Type *jed_find_syntax_table (char *name, int err) /*{{{*/
 {
    Syntax_Table_Type *table = Syntax_Tables;
    while (table != NULL)
      {
-	if (!strncmp (table->name, name, 15)) return table;
+	if (!strcmp (table->name, name)) return table;
 	table = table->next;
      }
    if (err) msg_error ("Syntax table undefined.");
@@ -1036,11 +1158,11 @@ Syntax_Table_Type *find_syntax_table (char *name, int err) /*{{{*/
 
 /*}}}*/
 
-void set_syntax_flags (char *name, int *flags) /*{{{*/
+static void set_syntax_flags (char *name, int *flags) /*{{{*/
 {
    Syntax_Table_Type *table;
 
-   table = find_syntax_table (name, 1);
+   table = jed_find_syntax_table (name, 1);
    if (table == NULL) return;
 
    table->flags |= *flags & 0xFF;
@@ -1048,7 +1170,8 @@ void set_syntax_flags (char *name, int *flags) /*{{{*/
 
 /*}}}*/
 
-void define_syntax (int *what, char *name) /*{{{*/
+/* UTF-8 WORD UNsafe */
+static void define_syntax (int *what, char *name) /*{{{*/
 {
    Syntax_Table_Type *table;
    int c2;
@@ -1056,7 +1179,7 @@ void define_syntax (int *what, char *name) /*{{{*/
    char *s1 = NULL, *s2 = NULL;
    unsigned char lut[256], *s;
 
-   table = find_syntax_table (name, 1);
+   table = jed_find_syntax_table (name, 1);
    if (table == NULL) return;
 
    switch (*what)
@@ -1064,16 +1187,23 @@ void define_syntax (int *what, char *name) /*{{{*/
       case '%':
 	if (SLang_pop_slstring (&s2)) break;
 	if (SLang_pop_slstring (&s1)) break;
-
+	
 	table->char_syntax[(unsigned char) *s1] |= COMMENT_SYNTAX;
 
 	if ((*s2 == 0) || (*s2 == '\n'))
 	  {
-	     SLang_free_slstring (table->eol_comment_start);
-
-	     table->eol_comment_start = s1;
-	     table->eol_comment_len = strlen (s1);
 	     SLang_free_slstring (s2);
+
+	     i = table->num_eol_comments;
+	     if (i == MAX_EOL_COMMENTS)
+	       {
+		  SLang_free_slstring (s1);
+		  return;
+	       }
+
+	     table->eol_comment_starts[i] = s1;
+	     table->eol_comment_lens[i] = strlen (s1);
+	     table->num_eol_comments = i + 1;
 	     return;
 	  }
 
@@ -1105,8 +1235,10 @@ void define_syntax (int *what, char *name) /*{{{*/
 	break;
       case '"':
 	if (SLang_pop_integer (&c2)) break;
+	if (table->num_string_chars == MAX_STRING_CHARS)
+	  break;
 	table->char_syntax[(unsigned char) c2] |= STRING_SYNTAX;
-	table->string_char = (unsigned char) c2;
+	table->string_chars[table->num_string_chars++] = (unsigned char) c2;
 	break;
 
       case '<':
@@ -1120,6 +1252,11 @@ void define_syntax (int *what, char *name) /*{{{*/
 	     table->char_syntax[(unsigned char) *(s2 + 1)] |= HTML_END_SYNTAX;
 	     s2 += 2;
 	  }
+	table->sgml_start_char = *s1;
+	if (*s1 != 0)
+	  table->sgml_stop_char = s1[1];
+	else
+	  table->sgml_stop_char = 0;
 	s2 = NULL;
 	break;
 
@@ -1200,12 +1337,12 @@ void define_syntax (int *what, char *name) /*{{{*/
 
 /*}}}*/
 
-void jed_set_fortran_comment_style (char *table_name, char *str)
+static void set_fortran_comment_style (char *table_name, char *str)
 {
    Syntax_Table_Type *table;
    int reverse;
 
-   if (NULL == (table = find_syntax_table (table_name, 1)))
+   if (NULL == (table = jed_find_syntax_table (table_name, 1)))
      return;
    
    reverse = 0;
@@ -1217,49 +1354,107 @@ void jed_set_fortran_comment_style (char *table_name, char *str)
    SLmake_lut(table->fortran_comment_chars, (unsigned char *) str, reverse);
 }
 
-void use_syntax_table (char *s) /*{{{*/
+static void use_syntax_table (char *s) /*{{{*/
 {
    Syntax_Table_Type *table;
 
    if ((s == NULL) || (*s == 0))
-     table = Default_Syntax_Table;
+     {
+	s = "";
+	table = Default_Syntax_Table;
+     }
    else
      {
-	table = find_syntax_table (s, 1);
+	table = jed_find_syntax_table (s, 1);
 	if (table == NULL) return;
      }
    CBuf->syntax_table = table;
+
+   (void) SLang_run_hooks ("use_syntax_table_hook", 1, s);
 }
 
 /*}}}*/
 
-void create_syntax_table (char *name) /*{{{*/
+/* Clears everything except for the name, and links to other tables */
+static void clear_syntax_table (Syntax_Table_Type *t)
+{
+   unsigned int i;
+   char *name;
+   Syntax_Table_Type *next;
+
+   for (i = 0; i < t->num_eol_comments; i++)
+     SLang_free_slstring (t->eol_comment_starts[i]);
+   SLang_free_slstring (t->comment_start);
+   SLang_free_slstring (t->comment_stop);
+   
+#if JED_HAS_DFA_SYNTAX
+   if (t->init_dfa_callback != NULL)
+     SLang_free_function (t->init_dfa_callback);
+   if (t->hilite != NULL)
+     jed_dfa_free_highlight_table (t->hilite);
+#endif
+   
+   for (i = 0; i < MAX_KEYWORD_TABLES; i++)
+     {
+	unsigned int j;
+	for (j = 0; j < MAX_KEYWORD_LEN; j++)
+	  {
+	     char *kwds = t->keywords[i][j];
+	     if (kwds != NULL)
+	       SLang_free_slstring (kwds);
+	  }
+     }
+
+   next = t->next;
+   name = t->name;
+   memset ((char *)t, 0, sizeof (Syntax_Table_Type));
+   t->next = next;
+   t->name = name;
+}
+
+static Syntax_Table_Type *allocate_syntax_table (char *name)
 {
    Syntax_Table_Type *table;
 
-   if (NULL != find_syntax_table (name, 0)) return;
+   if (NULL == (table = (Syntax_Table_Type *) jed_malloc0 (sizeof (Syntax_Table_Type))))
+     return NULL;
 
-   if (NULL == (table = (Syntax_Table_Type *) SLmalloc (sizeof (Syntax_Table_Type))))
+   if (NULL == (name = SLang_create_slstring (name)))
+     {
+	SLfree ((char *) table);
+	return NULL;
+     }
+   table->name = name;
+   return table;
+}
+
+
+static void create_syntax_table (char *name) /*{{{*/
+{
+   Syntax_Table_Type *table;
+
+   if (NULL != (table = jed_find_syntax_table (name, 0)))
+     {
+	clear_syntax_table (table);
+	return;
+     }
+
+   if (NULL == (table = allocate_syntax_table (name)))
      return;
 
-   SLMEMSET ((char *) table, 0, sizeof (Syntax_Table_Type));
    table->next = Syntax_Tables;
    Syntax_Tables = table;
-   strncpy (table->name, name, 15);
-   table->name[15] = 0;
 }
 
-/*}}}*/
 
 void init_syntax_tables (void) /*{{{*/
 {
    unsigned short *a;
    unsigned char *m;
 
-   Default_Syntax_Table = (Syntax_Table_Type *) SLmalloc (sizeof (Syntax_Table_Type));
+   Default_Syntax_Table = allocate_syntax_table ("DEFAULT");
    if (Default_Syntax_Table == NULL) return;
 
-   SLMEMSET ((char *) Default_Syntax_Table, 0, sizeof (Syntax_Table_Type));
    a = Default_Syntax_Table->char_syntax;
    m = Default_Syntax_Table->matching_delim;
 
@@ -1273,13 +1468,16 @@ void init_syntax_tables (void) /*{{{*/
 
 /*}}}*/
 
-void define_keywords (char *name, char *kwords, int *lenp, int *tbl_nump) /*{{{*/
+/* Currently this assumes byte-semantics.  It should be changed to assume character
+ * semantics.
+ */
+static void define_keywords (char *name, char *kwords, int *lenp, int *tbl_nump) /*{{{*/
 {
    char *kw;
    int len;
    int kwlen;
    unsigned int table_number = (unsigned int) *tbl_nump;
-   Syntax_Table_Type *table = find_syntax_table (name, 1);
+   Syntax_Table_Type *table = jed_find_syntax_table (name, 1);
 
    if (table == NULL) return;
 
@@ -1340,6 +1538,9 @@ static void syntax_parse_lines (Syntax_Table_Type *table, Line *l, unsigned int
 	     l->flags |= state;
 	     state = 0;
 	  }
+	if (((state == JED_LINE_IN_STRING0) || (state == JED_LINE_IN_STRING1))
+	    && (table->flags & SINGLE_LINE_STRINGS))
+	  state = 0;
 
 	l = l->next;
 
@@ -1366,9 +1567,14 @@ void jed_syntax_parse_buffer (int do_all)
    int is_narrow;
    Line *l;
    Syntax_Table_Type *table;
+   unsigned int undo_bit = 0;
+   SLang_Name_Type *color_region_hook = NULL;
+
+   if (CBuf->buffer_hooks != NULL)
+     color_region_hook = CBuf->buffer_hooks->color_region_hook;
 
    table = CBuf->syntax_table;
-   if (table == NULL)
+   if ((table == NULL) && (color_region_hook == NULL))
      {
 	CBuf->min_unparsed_line_num = 0;
 	CBuf->max_unparsed_line_num = 0;
@@ -1387,28 +1593,86 @@ void jed_syntax_parse_buffer (int do_all)
    is_narrow = (CBuf->narrow != NULL);
    if (is_narrow)
      {
+	/* yuk--- a major hack: turn off undo temporally */
+	undo_bit = CBuf->flags & UNDO_ENABLED;
+	CBuf->flags &= ~UNDO_ENABLED;
+
 	jed_push_narrow ();
 	jed_widen_whole_buffer (CBuf);
      }
 
+   CBuf->beg->flags &= ~JED_LINE_SYNTAX_BITS; /* 0.99-17.98 */
+
    if (do_all)
      {
 	min_line_num = 1;
 	max_line_num = Max_LineNum;
      }
-   n = (int) min_line_num;
+   
+   if (color_region_hook != NULL)
+     {
+	(void) SLang_start_arg_list ();
+	if ((0 == SLang_push_integer ((int) min_line_num))
+	    && (0 == SLang_push_integer ((int) max_line_num))
+	    && (0 == SLang_end_arg_list ()))
+	  {
+	     if (-1 == SLexecute_function (color_region_hook))
+	       CBuf->buffer_hooks->color_region_hook = NULL;
+	  }
+     }
+   else
+     {
+	n = (int) min_line_num;
 
-   push_spot ();
-   goto_line (&n);
-   l = CLine;
-   pop_spot ();
+	push_spot ();
+	goto_line (&n);
+	l = CLine;
+	pop_spot ();
 
-   syntax_parse_lines (table, l, 1 + (unsigned int) (max_line_num - min_line_num));
+	syntax_parse_lines (table, l, 1 + (unsigned int) (max_line_num - min_line_num));
+     }
 
    if (is_narrow)
-     jed_pop_narrow ();
+     {
+	jed_pop_narrow ();
+	CBuf->flags |= undo_bit;       /* hack */
+     }
 
    CBuf->min_unparsed_line_num = CBuf->max_unparsed_line_num = 0;
 }
 #endif
 
+static char *what_syntax_table (void)
+{
+   Syntax_Table_Type *s;
+
+   if ((NULL == (s = CBuf->syntax_table))
+       && (NULL == (s = Default_Syntax_Table)))
+     return NULL;
+   
+   return s->name;		       /* not thread safe */
+}
+
+static SLang_Intrin_Fun_Type Intrinsics [] = 
+{
+   MAKE_INTRINSIC("parse_to_point", parse_to_point, INT_TYPE, 0),
+   MAKE_INTRINSIC_SI("set_syntax_flags", set_syntax_flags, VOID_TYPE),
+   MAKE_INTRINSIC_IS("define_syntax", define_syntax, VOID_TYPE),
+   MAKE_INTRINSIC_S("use_syntax_table", use_syntax_table, VOID_TYPE),
+   MAKE_INTRINSIC_0("what_syntax_table", what_syntax_table, STRING_TYPE),
+   MAKE_INTRINSIC_S("create_syntax_table", create_syntax_table, VOID_TYPE),
+   MAKE_INTRINSIC_4("define_keywords_n", define_keywords, VOID_TYPE, STRING_TYPE, STRING_TYPE, INT_TYPE, INT_TYPE),
+   MAKE_INTRINSIC_SS("set_fortran_comment_chars", set_fortran_comment_style, SLANG_VOID_TYPE),
+   MAKE_INTRINSIC_I("find_matching_delimiter", find_matching_delimiter, INT_TYPE),
+   SLANG_END_INTRIN_FUN_TABLE
+};
+
+int jed_init_syntax (void)
+{
+#if JED_HAS_DFA_SYNTAX
+   if (-1 == jed_init_dfa_syntax ())
+     return -1;
+#endif
+
+   return SLadd_intrin_fun_table (Intrinsics, NULL);
+}
