@@ -1,5 +1,5 @@
 /* -*- mode: C; mode: fold; -*- */
-/* Copyright (c) 1992, 1998, 2000, 2002 John E. Davis
+/* Copyright (c) 1992, 1998, 2000, 2002, 2003, 2004, 2005, 2006 John E. Davis
  * This file is part of JED editor library source.
  *
  * You may distribute this file under the terms the GNU General Public
@@ -97,19 +97,19 @@ static void autoload_n (void) /*{{{*/
    char *file, *fun;
 
    if (SLang_pop_integer (&n)) return;
-   while (n > 0)
+   while ((n > 0) && (0 == SLang_get_error ()))
      {
 	n--;
-	if (SLang_pop_slstring (&file)) return;
+	if (SLang_pop_slstring (&file))
+	  return;
 	if (SLang_pop_slstring (&fun))
 	  {
 	     SLang_free_slstring (file);
 	     return;
 	  }
-	SLang_autoload (fun, file);
+	(void) SLang_autoload (fun, file);
 	SLang_free_slstring (file);
 	SLang_free_slstring (fun);
-	if (SLang_Error) break;
      }
 }
 
@@ -127,9 +127,12 @@ static void add_to_completion_n (void) /*{{{*/
 	if (SLang_pop_slstring (&fun))
 	  return;
 
-	add_to_completion (fun);
+	if (-1 == add_to_completion (fun))
+	  {
+	     SLang_free_slstring (fun);
+	     return;
+	  }
 	SLang_free_slstring (fun);
-	if (SLang_Error) break;
      }
 }
 
@@ -139,7 +142,6 @@ static void translate_region (void) /*{{{*/
 {
    SLang_Array_Type *a;
    char **strings;
-   int one = 1;
    int zero = 0;
    Mark *m;
 
@@ -151,7 +153,7 @@ static void translate_region (void) /*{{{*/
 
    if (-1 == SLang_pop_array_of_type (&a, SLANG_STRING_TYPE))
      return;
-   
+
    if (a->num_elements != 256)
      {
 	jed_verror ("translate_region: String_Type[256] required");
@@ -167,24 +169,23 @@ static void translate_region (void) /*{{{*/
 	char *s = strings [*(CLine->data + Point)];
 	if (s == NULL)
 	  {
-	     (void) forwchars (&one);
+	     (void) jed_right (1);
 	     continue;
 	  }
-	if (-1 == jed_ins_chars ((unsigned char *)s, strlen (s)))
+	if (-1 == jed_insert_nbytes ((unsigned char *)s, strlen (s)))
 	  break;
-	
-	if (-1 == jed_generic_deln (1))
+
+	if (-1 == jed_generic_del_nbytes (1))
 	  break;
      }
 
    /* Leave point at end of region */
-   pop_mark (&zero);
+   jed_pop_mark (0);
    SLang_free_array (a);
 }
 
 /*}}}*/
 
-
 static int write_region_cmd (char *file) /*{{{*/
 {
    int n = write_region (file);
@@ -237,7 +238,6 @@ static int rename_file (char *f1, char *f2) /*{{{*/
 
 /*}}}*/
 
-
 static void exit_error_cmd (char *msg, int *severity) /*{{{*/
 {
    exit_error (msg, *severity);
@@ -303,7 +303,7 @@ static void set_buffer_hook (void) /*{{{*/
      return;
 
    (void) jed_set_buffer_hook (CBuf, s, f);
-   
+
    SLang_free_slstring (s);
 }
 
@@ -323,10 +323,10 @@ static Buffer *pop_buffer (int nargs)
 
    if (nargs != SLang_Num_Function_Args)
      return CBuf;
-   
+
    if (-1 == SLang_pop_slstring (&s))
      return NULL;
-   
+
    if (NULL == (b = find_buffer (s)))
      jed_verror ("Unable to find a buffer named %s", s);
 
@@ -337,10 +337,12 @@ static Buffer *pop_buffer (int nargs)
 static void get_buffer_info(void) /*{{{*/
 {
    Buffer *b;
-   
+
    if (NULL == (b = pop_buffer (1)))
      return;
 
+   check_buffer (b);
+
    (void) SLang_push_string (b->file);
    (void) SLang_push_string (b->dir);
    (void) SLang_push_string (b->name);
@@ -353,7 +355,7 @@ static void set_buffer_info(char *file, char *dir, char *name, int *flagsp) /*{{
 {
    char *tmp;
    Buffer *b;
-   
+
    if (NULL == (b = pop_buffer (5)))
      return;
 
@@ -364,17 +366,17 @@ static void set_buffer_info(char *file, char *dir, char *name, int *flagsp) /*{{
    dir = tmp;
    fixup_dir (dir);
 
-   tmp = jed_expand_filename (dir);
+   tmp = jed_standardize_filename (dir);
    SLfree (dir);
    if (tmp == NULL)
      return;
 
    if ((file != b->file) || (tmp != b->dir))
      buffer_filename (b, tmp, file);
-   
+
    SLfree (tmp);
 
-   if (name != b->name) 
+   if (name != b->name)
      uniquely_name_buffer (b, name);
 
    jed_set_buffer_flags (b, (unsigned int) *flagsp);
@@ -384,10 +386,16 @@ static void set_buffer_info(char *file, char *dir, char *name, int *flagsp) /*{{
 
 static int intrin_what_mode(void) /*{{{*/
 {
-   char *m = CBuf->mode_string;
-   if (m == NULL) m = "";
+   Buffer *b;
+   char *m;
+
+   if (NULL == (b = pop_buffer (1)))
+     return -1;
+
+   if (NULL == (m = b->mode_string))
+     m = "";
    SLang_push_string (m);
-   return CBuf->modes;
+   return b->modes;
 }
 
 /*}}}*/
@@ -395,7 +403,7 @@ static int intrin_what_mode(void) /*{{{*/
 static void intrin_set_mode (int *flags)
 {
    char *mode;
-   
+
    if (-1 == SLang_pop_slstring (&mode))
      return;
 
@@ -407,7 +415,7 @@ static void intrin_set_mode (int *flags)
    CBuf->max_unparsed_line_num = Max_LineNum + CBuf->nup;
 #endif
 }
-   
+
 #ifdef REAL_UNIX_SYSTEM
 static char *get_termcap_string (char *cap) /*{{{*/
 {
@@ -421,14 +429,19 @@ static char *get_termcap_string (char *cap) /*{{{*/
 /*}}}*/
 #endif
 
-static void intrin_insert_char (int *ich)
+static void intrin_insert_char (unsigned long *chp)
 {
-   unsigned char ch = (unsigned char) *ich;
-   (void) jed_ins_chars (&ch, 1);
+   (void) jed_insert_wchar ((SLwchar_Type) *chp);
 }
+
+static void intrin_insert_byte (unsigned char *chp)
+{
+   (void) jed_insert_byte ((unsigned char)*chp);
+}
+
 static void del_intrinsic (void)
 {
-   (void) jed_del ();
+   (void) jed_del_wchar ();
 }
 
 static int file_changed_on_disk_cmd (char *file)
@@ -442,11 +455,7 @@ static int get_point_cmd (void)
 }
 static void set_point_cmd (int *p)
 {
-   if (*p >= CLine->len)
-     eol ();
-   else if (*p < 0)
-     Point = 0;
-   else Point = *p;
+   jed_position_point (CLine->data + *p);
 }
 
 static void expand_filename_cmd (char *s)
@@ -456,7 +465,7 @@ static void expand_filename_cmd (char *s)
 }
 
 static void insert_string (void)
-{   
+{
    char *s;
    SLang_BString_Type *bs;
    unsigned int len;
@@ -470,50 +479,19 @@ static void insert_string (void)
 	ret = jed_insert_string (s);
 	SLang_free_slstring (s);
 	break;
-	
+
       default:
 	if (-1 == SLang_pop_bstring (&bs))
 	  return;
 	if (NULL != (s = (char *)SLbstring_get_pointer (bs, &len)))
-	  ret = jed_ins_chars ((unsigned char *)s, len);
+	  ret = jed_insert_nbytes ((unsigned char *)s, len);
 	SLbstring_free (bs);
      }
-   
+
    if (ret == -1)
      jed_verror ("insert: insertion failed");
 }
 
-
-static void get_key_binding (void) 
-{
-   unsigned char *s = NULL;
-   char *str;
-   int type;
-
-   if (SLang_Num_Function_Args == 1)
-     {
-	if (-1 == SLang_pop_slstring (&str))
-	  return;
-	
-	s = (unsigned char *) SLang_process_keystring (str);
-	SLang_free_slstring (str);
-	
-	if (s == NULL)
-	  return;
-     }
-   
-   /* str is a statically allocated string */
-   type = jed_get_key_function (s, &str);
-   
-   (void) SLang_push_integer (type);
-   (void) SLang_push_string (str);
-}
-
-static void make_keymap_cmd (char *km)
-{
-   jed_copy_keymap_cmd (km, "global");
-}
-
 static int get_top_window_row (void)
 {
    return Top_Window_SY + 1;
@@ -539,7 +517,6 @@ static void _exit_intrin (int *status)
    (void) jed_quit_jed (*status);
 }
 
-
 static int jed_system (char *s) /*{{{*/
 {
    if (Jed_Secure_Mode)
@@ -547,7 +524,7 @@ static int jed_system (char *s) /*{{{*/
 	msg_error ("Access to shell denied.");
 	return -1;
      }
-   
+
 #ifdef IBMPC_SYSTEM
    return sys_System (s);
 #else
@@ -585,50 +562,135 @@ static void expand_link_intrin (char *s)
      (void) SLang_push_malloced_string (s);
 }
 
+static int right_intrinsic (int *np)
+{
+   if (*np < 0)
+     return 0;
+
+   return jed_right (*np);
+}
+
+static int left_intrinsic (int *np)
+{
+   if (*np < 0)
+     return 0;
+
+   return jed_left (*np);
+}
+
+static int up_intrinsic (int *np)
+{
+   if (*np < 0) return 0;
+   return jed_up ((unsigned int) *np);
+}
+
+static int down_intrinsic (int *np)
+{
+   if (*np < 0) return 0;
+   return jed_down ((unsigned int) *np);
+}
+
+static void pop_mark_intrin (int *go)
+{
+   jed_pop_mark (*go);
+}
+
+static void trim_intrinsic (void)
+{
+   (void) jed_trim_whitespace ();
+}
+
+static void skip_white_intrin (void)
+{
+   (void) jed_skip_whitespace ();
+}
+
+static void set_line_number_mode (int *statusp)
+{
+   int status = *statusp;
+   if (status < 0)
+     status = (CBuf->line_num_display_size == 0);
+
+   CBuf->line_num_display_size = (status > 0);
+}
+
+static void what_char_intrin (void)
+{
+   SLwchar_Type ch;
+   if (-1 == jed_what_char (&ch))
+     (void) SLang_push_integer (-(int)ch);
+   else
+     (void) SLclass_push_long_obj (SLANG_ULONG_TYPE, ch);
+}
+
+static int intrin_strwidth (char *s)
+{
+   char *smax;
+   
+   smax = s + strlen (s);
+#if SLANG_VERSION < 20000
+   /* This is technically not correct but slang-1 will not be supported for
+    * much longer.
+    */
+   return smax - s;
+#else
+   return (int) SLsmg_strwidth ((SLuchar_Type *)s, (SLuchar_Type *)smax);
+#endif
+}
+
+#if 0
+static int get_utf8_mode (void)
+{
+   return CBuf->local_vars.is_utf8;
+}
+
+static void set_utf8_mode (int *yes_no)
+{
+   CBuf->local_vars.is_utf8 = (*yes_no != 0);
+}
+#endif
+
 /* These pointers are necessary because the things they are assigned to
  * are character arrays.  Character arrays are not valid in MAKE_VARIABLE.
  */
 static char *Default_Status_Line_Ptr = Default_Status_Line;
-static char *Key_Buffer_Ptr = Jed_Key_Buffer;
 static char *Message_Buffer_Ptr = Message_Buffer;
-static char *Last_Kbd_Command_Str_Ptr = Last_Kbd_Command_String;
-static char *Current_Kbd_Command_Str_Ptr = Current_Kbd_Command_String;
 static char *Jed_Root_Dir_Ptr = Jed_Root_Dir;
 
 static SLang_Intrin_Fun_Type Jed_Intrinsics [] = /*{{{*/
 {
+#if 0
+   MAKE_INTRINSIC_0("get_utf8_mode", get_utf8_mode, INT_TYPE),
+   MAKE_INTRINSIC_I("set_utf8_mode", set_utf8_mode, VOID_TYPE),
+#endif
    MAKE_INTRINSIC_S("expand_symlink", expand_link_intrin, VOID_TYPE),
    MAKE_INTRINSIC_I("_set_point", set_point_cmd, VOID_TYPE),
    MAKE_INTRINSIC_0("_get_point", get_point_cmd, INT_TYPE),
-   MAKE_INTRINSIC_SS("setkey", set_key, VOID_TYPE),
    MAKE_INTRINSIC("_autoload", autoload_n, VOID_TYPE, 0),
-   MAKE_INTRINSIC("push_mark", push_mark,VOID_TYPE, 0),
+   MAKE_INTRINSIC("push_mark", jed_push_mark,VOID_TYPE, 0),
    MAKE_INTRINSIC("bol", bol,VOID_TYPE, 0),
    MAKE_INTRINSIC_0("insert", insert_string,VOID_TYPE),
-   MAKE_INTRINSIC_I("insert_char", intrin_insert_char, VOID_TYPE),
+   MAKE_INTRINSIC_1("insert_char", intrin_insert_char, VOID_TYPE, SLANG_ULONG_TYPE),
+   MAKE_INTRINSIC_1("insert_byte", intrin_insert_byte, VOID_TYPE, SLANG_UCHAR_TYPE),
    MAKE_INTRINSIC("eol", eol,VOID_TYPE, 0),
    MAKE_INTRINSIC_S("setbuf", set_buffer, VOID_TYPE),
    MAKE_INTRINSIC("_add_completion", add_to_completion_n, VOID_TYPE, 0),
    MAKE_INTRINSIC("del_region", delete_region, VOID_TYPE, 0),
    MAKE_INTRINSIC("bufsubstr", buffer_substring, VOID_TYPE, 0),
-   MAKE_INTRINSIC_I("right", forwchars,INT_TYPE),
-   MAKE_INTRINSIC_SSS("definekey", set_key_in_keymap, VOID_TYPE),
-   MAKE_INTRINSIC_I("left", backwchars,INT_TYPE),
+   MAKE_INTRINSIC_I("right", right_intrinsic, INT_TYPE),
+   MAKE_INTRINSIC_I("left", left_intrinsic, INT_TYPE),
    MAKE_INTRINSIC("whatbuf", what_buffer, STRING_TYPE, 0),
    MAKE_INTRINSIC("getbuf_info", get_buffer_info,  VOID_TYPE, 0),
-   MAKE_INTRINSIC("otherwindow", other_window, VOID_TYPE, 0),
    MAKE_INTRINSIC_S("is_internal", is_internal, INT_TYPE),
    MAKE_INTRINSIC_4("setbuf_info", set_buffer_info,  VOID_TYPE, STRING_TYPE, STRING_TYPE, STRING_TYPE, INT_TYPE),
-   MAKE_INTRINSIC_I("up", prevline, INT_TYPE),
-   MAKE_INTRINSIC_I("down", nextline,INT_TYPE),
+   MAKE_INTRINSIC_I("up", up_intrinsic, INT_TYPE),
+   MAKE_INTRINSIC_I("down", down_intrinsic, INT_TYPE),
    MAKE_INTRINSIC_S("call", jed_call_cmd, VOID_TYPE),
    MAKE_INTRINSIC("eob", eob,VOID_TYPE, 0),
-   MAKE_INTRINSIC_S("unsetkey", unset_key, VOID_TYPE),
    MAKE_INTRINSIC("bob", bob,VOID_TYPE, 0),
-   MAKE_INTRINSIC_S("looking_at", looking_at, INT_TYPE),
+   MAKE_INTRINSIC_S("looking_at", jed_looking_at, INT_TYPE),
    MAKE_INTRINSIC("del", del_intrinsic, VOID_TYPE, 0),
    MAKE_INTRINSIC("markp", markp, INT_TYPE, 0),
-   MAKE_INTRINSIC("nwindows", num_windows,INT_TYPE, 0),
    MAKE_INTRINSIC_S("add_completion", add_to_completion, VOID_TYPE),
    MAKE_INTRINSIC("what_column", calculate_column,INT_TYPE, 0),
    MAKE_INTRINSIC("eobp", eobp,INT_TYPE, 0),
@@ -636,7 +698,7 @@ static SLang_Intrin_Fun_Type Jed_Intrinsics [] = /*{{{*/
    MAKE_INTRINSIC_S("buffer_visible", jed_buffer_visible, INT_TYPE),
    MAKE_INTRINSIC("exit_jed", exit_jed_intrin, VOID_TYPE, 0),
    MAKE_INTRINSIC_S("extract_filename", extract_file, STRING_TYPE),
-   MAKE_INTRINSIC("trim", trim_whitespace,VOID_TYPE, 0),
+   MAKE_INTRINSIC_0("trim", trim_intrinsic, VOID_TYPE),
    MAKE_INTRINSIC_S("pop2buf", pop_to_buffer, VOID_TYPE),
    MAKE_INTRINSIC_S("pop2buf_whatbuf", pop_to_buffer, STRING_TYPE),
 #ifndef VMS
@@ -646,19 +708,16 @@ static SLang_Intrin_Fun_Type Jed_Intrinsics [] = /*{{{*/
    MAKE_INTRINSIC_S("insbuf", insert_buffer_name, VOID_TYPE),
    MAKE_INTRINSIC("bolp", bolp,INT_TYPE, 0),
    MAKE_INTRINSIC("beep", jed_beep, VOID_TYPE, 0),
-   MAKE_INTRINSIC("onewindow", one_window,VOID_TYPE, 0),
    MAKE_INTRINSIC("pop_spot", pop_spot,VOID_TYPE, 0),
    MAKE_INTRINSIC("push_spot", push_spot,VOID_TYPE, 0),
    MAKE_INTRINSIC_S("sw2buf", switch_to_buffer_cmd, VOID_TYPE),
    MAKE_INTRINSIC_S("tt_send", do_tt_write_string, VOID_TYPE),
    MAKE_INTRINSIC("eolp", eolp,INT_TYPE, 0),
-   MAKE_INTRINSIC("what_keymap", what_keymap, STRING_TYPE, 0),
    MAKE_INTRINSIC_S("find_file", find_file_in_window, INT_TYPE),
    MAKE_INTRINSIC_SI("set_status_line", set_status_format, VOID_TYPE),
    MAKE_INTRINSIC_S("bury_buffer", bury_buffer, VOID_TYPE),
    MAKE_INTRINSIC("dupmark", dup_mark, INT_TYPE, 0),
    MAKE_INTRINSIC("erase_buffer", erase_buffer, VOID_TYPE, 0),
-   MAKE_INTRINSIC_I("window_info", window_size_intrinsic, INT_TYPE),
 #ifndef SIXTEEN_BIT_SYSTEM
    MAKE_INTRINSIC_I("copy_region_to_kill_array", copy_region_to_kill_array, VOID_TYPE),
    MAKE_INTRINSIC_I("append_region_to_kill_array", append_region_to_kill_array, VOID_TYPE),
@@ -668,35 +727,28 @@ static SLang_Intrin_Fun_Type Jed_Intrinsics [] = /*{{{*/
    MAKE_INTRINSIC_I("goto_column", goto_column, VOID_TYPE),
    MAKE_INTRINSIC_I("goto_column_best_try", goto_column1, INT_TYPE),
    MAKE_INTRINSIC_I("goto_line", goto_line,VOID_TYPE),
-   MAKE_INTRINSIC_I("pop_mark", pop_mark, VOID_TYPE),
+   MAKE_INTRINSIC_I("pop_mark", pop_mark_intrin, VOID_TYPE),
    MAKE_INTRINSIC_SSS("read_mini", mini_read, VOID_TYPE),
    MAKE_INTRINSIC_S("file_status", file_status, INT_TYPE),
-   MAKE_INTRINSIC("skip_white", skip_whitespace, VOID_TYPE, 0),
+   MAKE_INTRINSIC_0("skip_white", skip_white_intrin, VOID_TYPE),
    MAKE_INTRINSIC("bobp", bobp,INT_TYPE, 0),
    MAKE_INTRINSIC_S("flush", flush_message, VOID_TYPE),
    MAKE_INTRINSIC_I("input_pending", input_pending, INT_TYPE),
    MAKE_INTRINSIC_I("usleep", jed_pause, VOID_TYPE),
    MAKE_INTRINSIC_S("insert_file",  insert_file, INT_TYPE),
-   MAKE_INTRINSIC_S("keymap_p",  keymap_p, INT_TYPE),
-   MAKE_INTRINSIC("what_char", what_char, INT_TYPE, 0),
+   MAKE_INTRINSIC_0("what_char", what_char_intrin, VOID_TYPE),
    MAKE_INTRINSIC_I("recenter", recenter, VOID_TYPE),
    MAKE_INTRINSIC_S("bufferp", bufferp, INT_TYPE),
-   MAKE_INTRINSIC("get_key_binding", get_key_binding, VOID_TYPE, 0),
-   MAKE_INTRINSIC_S("dump_bindings", dump_bindings, VOID_TYPE),
-   MAKE_INTRINSIC_SS("undefinekey", unset_key_in_keymap, VOID_TYPE),
    MAKE_INTRINSIC_I("update", update_cmd, VOID_TYPE),
    MAKE_INTRINSIC_I("update_sans_update_hook", update_sans_update_hook_cmd, VOID_TYPE),
    MAKE_INTRINSIC("skip_word_chars", skip_word_chars, VOID_TYPE, 0),
    MAKE_INTRINSIC("skip_non_word_chars", skip_non_word_chars, VOID_TYPE, 0),
    MAKE_INTRINSIC("bskip_word_chars", bskip_word_chars, VOID_TYPE, 0),
    MAKE_INTRINSIC("bskip_non_word_chars", bskip_non_word_chars, VOID_TYPE, 0),
-   MAKE_INTRINSIC_S("which_key", which_key, INT_TYPE),
    MAKE_INTRINSIC_I("whitespace", insert_whitespace,VOID_TYPE),
-   MAKE_INTRINSIC("enlargewin", enlarge_window,VOID_TYPE, 0),
-   MAKE_INTRINSIC("splitwindow", split_window,VOID_TYPE, 0),
    MAKE_INTRINSIC_SS("file_time_compare", file_time_cmp, INT_TYPE),
-   MAKE_INTRINSIC_I("xform_region", transform_region, VOID_TYPE),
-   MAKE_INTRINSIC_S("skip_chars", skip_chars, VOID_TYPE),
+   /* MAKE_INTRINSIC_I("xform_region", transform_region, VOID_TYPE), */
+   MAKE_INTRINSIC_S("skip_chars", jed_skip_chars, VOID_TYPE),
    MAKE_INTRINSIC_I("set_file_translation", set_file_trans, VOID_TYPE),
 #if defined (__unix__) || defined(__WIN32__) || (defined (__os2__) && !defined(__WATCOMC__))
 # if (!defined(__GO32__) && !defined(__WATCOMC__)) || defined(__QNX__)
@@ -709,7 +761,7 @@ static SLang_Intrin_Fun_Type Jed_Intrinsics [] = /*{{{*/
    MAKE_INTRINSIC("autosaveall", auto_save_all, VOID_TYPE, 0),
    MAKE_INTRINSIC("backward_paragraph", backward_paragraph, VOID_TYPE, 0),
    MAKE_INTRINSIC("blank_rect", blank_rectangle, VOID_TYPE, 0),
-   MAKE_INTRINSIC_S("bskip_chars", bskip_chars, VOID_TYPE),
+   MAKE_INTRINSIC_S("bskip_chars", jed_bskip_chars, VOID_TYPE),
    MAKE_INTRINSIC("buffer_list", make_buffer_list,  VOID_TYPE, 0),
    MAKE_INTRINSIC_I("check_region", check_region, VOID_TYPE),
    MAKE_INTRINSIC("copy_rect", copy_rectangle, VOID_TYPE, 0),
@@ -722,12 +774,9 @@ static SLang_Intrin_Fun_Type Jed_Intrinsics [] = /*{{{*/
    MAKE_INTRINSIC_0("check_buffers", check_buffers, SLANG_VOID_TYPE),
    MAKE_INTRINSIC_S("file_changed_on_disk", file_changed_on_disk_cmd, INT_TYPE),
    MAKE_INTRINSIC("forward_paragraph", forward_paragraph, VOID_TYPE, 0),
-   MAKE_INTRINSIC("getkey", jed_getkey, INT_TYPE, 0),
    MAKE_INTRINSIC("indent_line", indent_line, VOID_TYPE, 0),
    MAKE_INTRINSIC("insert_rect", insert_rectangle, VOID_TYPE, 0),
    MAKE_INTRINSIC("kill_rect", kill_rectangle, VOID_TYPE, 0),
-   MAKE_INTRINSIC_S("make_keymap", make_keymap_cmd, VOID_TYPE),
-   MAKE_INTRINSIC_SS("copy_keymap", jed_copy_keymap_cmd, VOID_TYPE),
    MAKE_INTRINSIC_II("map_input", map_character, VOID_TYPE),
    MAKE_INTRINSIC("narrow_to_region", narrow_to_region, VOID_TYPE, 0),
    MAKE_INTRINSIC("narrow", narrow_to_lines, VOID_TYPE, 0),
@@ -739,9 +788,7 @@ static SLang_Intrin_Fun_Type Jed_Intrinsics [] = /*{{{*/
    MAKE_INTRINSIC_4("read_with_completion", read_object_with_completion, VOID_TYPE, STRING_TYPE, STRING_TYPE, STRING_TYPE, INT_TYPE),
    MAKE_INTRINSIC_I("set_abort_char", jed_set_abort_char, VOID_TYPE),
    MAKE_INTRINSIC("suspend", sys_spawn_cmd, VOID_TYPE, 0),
-   MAKE_INTRINSIC_I("ungetkey", ungetkey, VOID_TYPE),
    MAKE_INTRINSIC_S("buffer_keystring", do_buffer_keystring, VOID_TYPE),
-   MAKE_INTRINSIC_S("use_keymap",  use_keymap, VOID_TYPE),
 #ifndef IBMPC_SYSTEM
    MAKE_INTRINSIC("w132", screen_w132, VOID_TYPE, 0),
    MAKE_INTRINSIC("w80", screen_w80, VOID_TYPE, 0),
@@ -749,10 +796,9 @@ static SLang_Intrin_Fun_Type Jed_Intrinsics [] = /*{{{*/
    MAKE_INTRINSIC("what_mode", intrin_what_mode, INT_TYPE, 0),
    MAKE_INTRINSIC("widen", widen, VOID_TYPE, 0),
    MAKE_INTRINSIC("widen_region", widen_region, VOID_TYPE, 0),
-   MAKE_INTRINSIC("window_line", window_line, INT_TYPE, 0),
    MAKE_INTRINSIC_S("write_buffer", jed_save_buffer_as, INT_TYPE),
    MAKE_INTRINSIC_S("write_region_to_file", write_region_cmd, INT_TYPE),
-   MAKE_INTRINSIC("count_chars", count_chars, VOID_TYPE, 0),
+   MAKE_INTRINSIC("count_chars", jed_count_chars, VOID_TYPE, 0),
    MAKE_INTRINSIC_S("get_yes_no", jed_get_yes_no, INT_TYPE),
    MAKE_INTRINSIC_S("get_y_or_n", jed_get_y_n, INT_TYPE),
    MAKE_INTRINSIC_S("get_mini_response", jed_get_mini_response, INT_TYPE),
@@ -782,8 +828,6 @@ static SLang_Intrin_Fun_Type Jed_Intrinsics [] = /*{{{*/
    MAKE_INTRINSIC("user_mark_buffer", user_mark_buffer, STRING_TYPE, 0),
    MAKE_INTRINSIC("is_user_mark_in_narrow", jed_is_user_mark_in_narrow, INT_TYPE, 0),
 
-
-
 #if JED_HAS_LINE_MARKS
    MAKE_INTRINSIC_I("create_line_mark", jed_create_line_mark, VOID_TYPE),
 #endif
@@ -803,6 +847,9 @@ static SLang_Intrin_Fun_Type Jed_Intrinsics [] = /*{{{*/
    MAKE_INTRINSIC_I("kill_process", jed_close_process, VOID_TYPE),
    MAKE_INTRINSIC_IS("send_process", jed_send_process, VOID_TYPE),
    MAKE_INTRINSIC_I("open_process", jed_open_process, INT_TYPE),
+#ifdef REAL_UNIX_SYSTEM
+   MAKE_INTRINSIC_I("open_process_pipe", jed_open_process_pipe, INT_TYPE),
+#endif
    MAKE_INTRINSIC_I("process_mark", jed_get_process_mark, VOID_TYPE),
    MAKE_INTRINSIC_0("set_process", jed_set_process, VOID_TYPE),
    MAKE_INTRINSIC_I("send_process_eof", jed_send_process_eof, VOID_TYPE),
@@ -833,7 +880,7 @@ static SLang_Intrin_Fun_Type Jed_Intrinsics [] = /*{{{*/
    MAKE_INTRINSIC_SS("vms_get_help", vms_get_help,VOID_TYPE),
    MAKE_INTRINSIC_SS("vms_send_mail", vms_send_mail, INT_TYPE),
 #endif
-   
+
 #ifdef REAL_UNIX_SYSTEM
    MAKE_INTRINSIC_I("enable_flow_control", enable_flow_control, VOID_TYPE),
    /* MAKE_INTRINSIC_S("get_passwd_info", get_passwd_info, VOID_TYPE), */
@@ -844,14 +891,15 @@ static SLang_Intrin_Fun_Type Jed_Intrinsics [] = /*{{{*/
    MAKE_INTRINSIC_S("IsHPFSFileSystem", IsHPFSFileSystem, INT_TYPE),
 #endif
    MAKE_INTRINSIC_0("TOP_WINDOW_ROW", get_top_window_row, INT_TYPE),
-   
+
    MAKE_INTRINSIC_S("run_program", run_program, INT_TYPE),
-   MAKE_INTRINSIC(NULL, NULL, 0, 0)
+   MAKE_INTRINSIC_I("set_line_number_mode", set_line_number_mode, INT_TYPE),
+   MAKE_INTRINSIC_S("strwidth", intrin_strwidth, INT_TYPE),
+   SLANG_END_INTRIN_FUN_TABLE
 };
 
 /*}}}*/
 
-
 #ifndef IBMPC_SYSTEM
 static int Obsolete_Int_Variable;
 #endif
@@ -859,10 +907,10 @@ static SLang_Intrin_Var_Type Jed_Variables [] =
 {
    MAKE_VARIABLE("Status_Line_String", &Default_Status_Line_Ptr, STRING_TYPE, 1),
    MAKE_VARIABLE("MINIBUFFER_ACTIVE", &MiniBuffer_Active, INT_TYPE, 1),
-   MAKE_VARIABLE("DISPLAY_EIGHT_BIT", &SLsmg_Display_Eight_Bit, INT_TYPE, 0),
    MAKE_VARIABLE("USE_TABS", &Jed_Use_Tabs, INT_TYPE, 0),
    MAKE_VARIABLE("TAB_DEFAULT", &Jed_Tab_Default, INT_TYPE, 0),
-   MAKE_VARIABLE("CASE_SEARCH", &Case_Sensitive, INT_TYPE, 0),
+   MAKE_VARIABLE("CASE_SEARCH", &Buffer_Local.case_search, INT_TYPE, 0),
+   MAKE_VARIABLE("CASE_SEARCH_DEFAULT", &Jed_Case_Search_Default, INT_TYPE, 0),
    MAKE_VARIABLE("REPLACE_PRESERVE_CASE_INTERNAL", &Replace_Preserve_Case, INT_TYPE, 0),
    /* Prototype: Integer REPLACE_PRESERVE_CASE_INTERNAL = 0;
     * This variable affects the @replace@ function.  If the length of the
@@ -879,7 +927,6 @@ static SLang_Intrin_Var_Type Jed_Variables [] =
    MAKE_VARIABLE("WRAP_INDENTS", &Indented_Text_Mode, INT_TYPE, 0),
    MAKE_VARIABLE("IGNORE_BEEP", &tt_Ignore_Beep, INTP_TYPE, 0),
    MAKE_VARIABLE("ADD_NEWLINE", &Require_Final_Newline, INT_TYPE, 0),
-   MAKE_VARIABLE("LASTKEY", &Key_Buffer_Ptr, STRING_TYPE, 1),
    MAKE_VARIABLE("DISPLAY_TIME", &Display_Time, INT_TYPE, 0),
    MAKE_VARIABLE("WANT_EOB", &Want_Eob, INT_TYPE, 0),
    MAKE_VARIABLE("WRAP", &Jed_Wrap_Column, INT_TYPE, 0),
@@ -887,7 +934,6 @@ static SLang_Intrin_Var_Type Jed_Variables [] =
    MAKE_VARIABLE("DEC_8BIT_HACK", &DEC_8Bit_Hack, INT_TYPE, 0),
    MAKE_VARIABLE("BATCH", &Batch, INT_TYPE, 1),
    MAKE_VARIABLE("TAB", &Buffer_Local.tab, INT_TYPE, 0),
-   MAKE_VARIABLE("LAST_CHAR", &SLang_Last_Key_Char, INT_TYPE, 0),
    MAKE_VARIABLE("MAX_HITS", &Jed_Max_Hits, INT_TYPE, 0),
    MAKE_VARIABLE("POINT", &Point, INT_TYPE, 0),
    MAKE_VARIABLE("MESSAGE_BUFFER", &Message_Buffer_Ptr, STRING_TYPE, 1),
@@ -902,9 +948,8 @@ static SLang_Intrin_Var_Type Jed_Variables [] =
    MAKE_VARIABLE("WANT_SYNTAX_HIGHLIGHT", &Wants_Syntax_Highlight, INT_TYPE, 0),
    MAKE_VARIABLE("DEFINING_MACRO", &Defining_Keyboard_Macro, INT_TYPE, 1),
    MAKE_VARIABLE("EXECUTING_MACRO", &Executing_Keyboard_Macro, INT_TYPE, 1),
-   MAKE_VARIABLE("LAST_KBD_COMMAND", &Last_Kbd_Command_Str_Ptr, STRING_TYPE, 1),
-   MAKE_VARIABLE("CURRENT_KBD_COMMAND", &Current_Kbd_Command_Str_Ptr, STRING_TYPE, 1),
    MAKE_VARIABLE("DOLLAR_CHARACTER", &Jed_Dollar, INT_TYPE, 0),
+   MAKE_VARIABLE("Highlight_Trailing_Whitespace", &Jed_Highlight_Trailing_WS, INT_TYPE, 0),
    MAKE_VARIABLE("_jed_version", &Jed_Version_Number, INT_TYPE, 1),
    MAKE_VARIABLE("_jed_version_string", &Jed_Version_String, STRING_TYPE, 1),
    MAKE_VARIABLE("_jed_secure_mode", &Jed_Secure_Mode, INT_TYPE, 1),
@@ -931,7 +976,7 @@ static SLang_Intrin_Var_Type Jed_Variables [] =
 # endif
    MAKE_VARIABLE("ALT_CHAR", &PC_Alt_Char, INT_TYPE, 0),
 #endif				       /* IBMPC_SYSTEM */
-     
+
 #ifdef REAL_UNIX_SYSTEM
    MAKE_VARIABLE("BACKUP_BY_COPYING", &Jed_Backup_By_Copying, INT_TYPE, 0),
 #endif
@@ -955,11 +1000,19 @@ int init_jed_intrinsics (void) /*{{{*/
        || (-1 == SLadd_intrin_fun_table(JedLine_Intrinsics, NULL))
 #endif
        || (-1 == jed_init_color_intrinsics ())
+       || (-1 == jed_init_keymap_intrinsics ())
+       || (-1 == jed_init_window_intrinsics ())
        )
      return -1;
-   
+
+   /* made here for win32 dll support! */
+   if ((-1 == SLadd_intrinsic_variable ("DISPLAY_EIGHT_BIT", (VOID_STAR)&SLsmg_Display_Eight_Bit, INT_TYPE, 0))
+       || (-1 == SLadd_intrinsic_variable ("LAST_CHAR", (VOID_STAR)&SLang_Last_Key_Char, INT_TYPE, 0))
+       )
+     return -1;
+
    /* overload default slang version of system */
-   SLadd_intrinsic_function ("system", (FVOID_STAR) jed_system, 
+   SLadd_intrinsic_function ("system", (FVOID_STAR) jed_system,
 			     SLANG_INT_TYPE, 1, SLANG_STRING_TYPE);
 #if JED_HAS_DISPLAY_TABLE
    SLang_add_intrinsic_array ("Display_Map",   /* slang name */
@@ -978,7 +1031,7 @@ int init_jed_intrinsics (void) /*{{{*/
 			      256);    /* number of elements in X direction */
 #endif
 #ifdef __MSDOS__
-   /* SLang does not define this symbol if WIN32 was used.  However, as 
+   /* SLang does not define this symbol if WIN32 was used.  However, as
     * far as JED is concerned, WIN32 is MSDOS because of the file system.
     */
    SLdefine_for_ifdef ("MSDOS");
@@ -1000,7 +1053,7 @@ int init_jed_intrinsics (void) /*{{{*/
    SLdefine_for_ifdef("HAS_BLOCAL_VAR");
 #endif
 
-   if (SLang_Error) return -1;
+   if (SLang_get_error ()) return -1;
 
    return 0;
 }
