@@ -1,4 +1,4 @@
-/* Copyright (c) 1992, 1998 John E. Davis
+/* Copyright (c) 1992, 1998, 2000, 2005, 2006 John E. Davis
  * This file is part of JED editor library source.
  *
  * You may distribute this file under the terms the GNU General Public
@@ -27,13 +27,18 @@ extern void read_process_input (int);
 extern int jed_close_process (int *);
 extern int jed_send_process (int *, char *);
 extern int jed_open_process (int *);
+extern int jed_open_process_pipe (int *);
+
 extern void jed_get_child_status (void);
 extern void jed_kill_process (int);
 extern void jed_get_process_mark (int *);
-extern void jed_set_process (int *, char *, char *);
+extern void jed_set_process (void);
 extern void jed_send_process_eof (int *);
 extern void get_process_input (int *);
 extern int jed_signal_process (int *, int *);
+extern int jed_signal_fg_process (int *, int *);
+extern int jed_processes_ok_to_exit (void);
+extern void jed_query_process_at_exit (int *, int *);
 
 #ifdef REAL_UNIX_SYSTEM
 extern void jed_block_child_signal (int);
