@@ -1,5 +1,5 @@
 /* -*- mode: C; mode: fold; -*- */
-/* Copyright (c) 1992, 1998, 2000, 2002 John E. Davis
+/* Copyright (c) 1992, 1998, 2000, 2002, 2003, 2004, 2005, 2006 John E. Davis
  * This file is part of JED editor library source.
  *
  * You may distribute this file under the terms the GNU General Public
@@ -44,6 +44,10 @@
 
 /*}}}*/
 
+#if (SLANG_VERSION < 20003) && defined(SLKEY_F_SLANG)
+# undef SLKEY_F_SLANG
+#endif
+
 void (*X_Define_Keys_Hook)(SLKeyMap_List_Type *);
 int (*X_Set_Abort_Char_Hook) (unsigned char);
 unsigned char Jed_Abort_Char = 7;		       /* ^G */
@@ -192,8 +196,8 @@ SLKeymap_Function_Type Jed_Functions[] = /*{{{*/
      {"redraw", redraw},
       /* {"replace", replace}, */
      {"save_some_buffers", save_some_buffers},
-     {"scroll_left", scroll_left},
-     {"scroll_right", scroll_right},
+     {"scroll_left", jed_scroll_left_cmd},
+     {"scroll_right", jed_scroll_right_cmd},
        /* {"search_backward", search_backward_cmd},
 	  {"search_forward", search_forward_cmd}, */
      {"set_mark_cmd", set_mark_cmd},
@@ -202,7 +206,7 @@ SLKeymap_Function_Type Jed_Functions[] = /*{{{*/
      {"sys_spawn_cmd", sys_spawn_cmd},
      {"text_smart_quote", text_smart_quote},
 /*       {"transpose_lines", transpose_lines}, */
-     {"trim_whitespace", trim_whitespace},
+     {"trim_whitespace", jed_trim_whitespace},
      {"undo", undo},
      {"widen", widen},
      {"widen_region", widen_region},
@@ -220,14 +224,15 @@ SLKeyMap_List_Type *Global_Map;
 
 int kbd_quit(void) /*{{{*/
 {
-   int sle = SLang_Error;
+   int sle = SLang_get_error ();
    SLKeyBoard_Quit = 1;
 #if JED_HAS_TTY_MENUS
    if (Jed_Menus_Active)
      jed_exit_menu_bar ();
 #endif
    msg_error("Quit!");
-   if (Ignore_User_Abort) SLang_Error = sle;
+   if (Ignore_User_Abort) 
+     SLang_set_error (sle);
 #ifdef UNDO_HAS_REDO
    set_current_undo ();
 #endif
@@ -322,8 +327,8 @@ void init_keymaps(void) /*{{{*/
    SLkm_define_key ("^[OC", (FVOID_STAR) next_char_cmd, Global_Map);
    SLkm_define_key ("^[[6~", (FVOID_STAR) pagedown_cmd, Global_Map);
    SLkm_define_key ("^[[5~", (FVOID_STAR) pageup_cmd, Global_Map);
-   SLkm_define_key ("^K^[[C", (FVOID_STAR) scroll_left, Global_Map);
-   SLkm_define_key ("^K^[[D", (FVOID_STAR) scroll_right, Global_Map);
+   SLkm_define_key ("^K^[[C", (FVOID_STAR) jed_scroll_left_cmd, Global_Map);
+   SLkm_define_key ("^K^[[D", (FVOID_STAR) jed_scroll_right_cmd, Global_Map);
    SLkm_define_key ("^K^[[A", (FVOID_STAR) bob, Global_Map);
    SLkm_define_key ("^K^[[B", (FVOID_STAR)eob, Global_Map);
 #if HAS_MOUSE
@@ -395,7 +400,7 @@ void init_keymaps(void) /*{{{*/
    SLkm_define_key ("^[Q", (FVOID_STAR) text_format_paragraph, Global_Map);
    SLkm_define_key ("^[S", (FVOID_STAR) center_line, Global_Map);
    SLkm_define_key ("^[X", (FVOID_STAR) evaluate_cmd, Global_Map);
-   SLkm_define_key ("^[\\", (FVOID_STAR) trim_whitespace, Global_Map);
+   SLkm_define_key ("^[\\", (FVOID_STAR) jed_trim_whitespace, Global_Map);
    SLkm_define_key ("^\\", (FVOID_STAR) goto_match, Global_Map);
    SLkm_define_key ("`", (FVOID_STAR) quoted_insert, Global_Map);
    
@@ -456,7 +461,7 @@ static int do_macro_string (char *str, int repeat)
 
 	Read_This_Character = str;
    
-	while ((SLang_Error == 0)
+	while ((SLang_get_error () == 0)
 	       && (SLKeyBoard_Quit == 0)
 	       && (Read_This_Character != NULL))
 	  jed_do_key();
@@ -482,13 +487,14 @@ static int key_interpret (SLang_Key_Type *key) /*{{{*/
    strcpy (Last_Kbd_Command_String, Current_Kbd_Command_String);
    Jed_This_Key_Function = key->f.f;
 
-   if (key->type == SLKEY_F_INTRINSIC)
+   switch (key->type)
      {
+      case SLKEY_F_INTRINSIC:
 	set_current_kbd_command (Jed_Key_Buffer);
 	ret = (key->f.f) ();
-     }
-   else
-     {
+	break;
+	
+      case SLKEY_F_INTERPRET:
 	str = key->f.s;
 	set_current_kbd_command (str);
 
@@ -513,10 +519,21 @@ static int key_interpret (SLang_Key_Type *key) /*{{{*/
 	else if ((*str == '.')
 		 || !SLang_execute_function(str))
 	  SLang_load_string(str);
-	
+
 	ret = 1;
+	break;
+
+#ifdef SLKEY_F_SLANG
+      case SLKEY_F_SLANG:
+	ret = SLexecute_function (key->f.slang_fun);
+	break;
+#endif
+
+      default:
+	SLang_verror (SL_NOT_IMPLEMENTED, "Keymap key-type %d is not supported", (int)key->type);
+	return -1;
      }
-   
+
    Last_Key_Function = Jed_This_Key_Function;
    return(ret);
 }
@@ -545,10 +562,21 @@ static char *lookup_key_function_string (SLang_Key_Type *key)
    if (key == NULL)
      return NULL;
 
-   if (key->type != SLKEY_F_INTRINSIC)
-     return key->f.s;
-   
-   return find_function_string (key->f.f);
+   switch (key->type)
+     {
+      case SLKEY_F_INTRINSIC:
+	return find_function_string (key->f.f);
+	
+      case SLKEY_F_INTERPRET:
+	return key->f.s;
+	
+#ifdef SLKEY_F_SLANG
+      case SLKEY_F_SLANG:
+	return key->f.slang_fun->name;
+#endif
+      default:
+	return "";
+     }
 }
 
 
@@ -592,7 +620,7 @@ static int do_key (void) /*{{{*/
 	  {
 	     repeat--;
 
-	     if (SLKeyBoard_Quit || SLang_Error || (Repeat_Factor == NULL)) 
+	     if (SLKeyBoard_Quit || SLang_get_error () || (Repeat_Factor == NULL)) 
 	       break;
 
 	     key_interpret (key);
@@ -615,7 +643,7 @@ static int do_key (void) /*{{{*/
 int jed_do_key (void)
 {
    int ret = do_key ();
-   if (SLang_Error)
+   if (SLang_get_error ())
      {
 	jed_abort_keyboard_macro ();
 	Read_This_Character = NULL;
@@ -690,6 +718,14 @@ void jed (void) /*{{{*/
       one more hook to just to make sure  all things are go. */
    char *buffer_name;
 
+   if (SLang_get_error ())
+     {
+	update (NULL, 1, 1, 0);
+	SLang_restart (1);
+	SLang_set_error (0);
+	jed_sleep (2);
+     }
+
 #if JED_HAS_TTY_MENUS
    jed_notify_menu_buffer_changed ();
 #endif
@@ -722,11 +758,23 @@ void jed (void) /*{{{*/
 
    while(1)
      {
+	int had_err = 0;
 	Suspend_Screen_Update = 0;
-
+	
 	do_jed();
+	if (SLang_get_error ())
+	  {
+	     update (NULL, 1, 0, 0);
+	     SLang_restart (1);
+	     SLang_set_error (0);
+	     had_err = 1;
+	  }
+
 	if (!SLKeyBoard_Quit && (CBuf->flags & BUFFER_MODIFIED)
 	    && (!Cursor_Motion)) CBuf->hits += 1;
+	
+	SLKeyBoard_Quit = 0;
+
 	if (CBuf->hits > Jed_Max_Hits)
 	  {
 	     auto_save_buffer(CBuf);
@@ -741,7 +789,7 @@ void jed (void) /*{{{*/
 	    || Suspend_Screen_Update ||
 	    (Mini_Ghost) || (*Message_Buffer) || (*Error_Buffer))
 	  {
-	     update((Line *) NULL, 0, 0, 1);
+	     update((Line *) NULL, 0, had_err, 1);
 	  }
      }
 }
@@ -792,7 +840,7 @@ int digit_arg(void) /*{{{*/
 
 /*}}}*/
 
-int which_key (char *f) /*{{{*/
+static int which_key (char *f) /*{{{*/
 {
    int num = 0, i;
    SLang_Key_Type *key, *key_root;
@@ -881,9 +929,9 @@ static void dump_this_binding (SLang_Key_Type *key) /*{{{*/
 	     n = 2;
 	  }
 	(void) jed_quick_insert((unsigned char *) str, n);
-	(void) jed_ins(' ');
+	(void) jed_insert_byte (' ');
      }
-   (void) jed_ins_char_n_times('\t', 3);
+   (void) jed_insert_wchar_n_times('\t', 3);
    
    fun = lookup_key_function_string (key);
    
@@ -894,7 +942,7 @@ static void dump_this_binding (SLang_Key_Type *key) /*{{{*/
 
 /*}}}*/
 
-void dump_bindings(char *map) /*{{{*/
+static void dump_bindings(char *map) /*{{{*/
 {
    int i;
    SLang_Key_Type *next, *key_root;
@@ -927,7 +975,7 @@ void dump_bindings(char *map) /*{{{*/
 
 /*}}}*/
 
-void use_keymap(char *name) /*{{{*/
+static void use_keymap(char *name) /*{{{*/
 {
    SLKeyMap_List_Type *map;
    
@@ -941,7 +989,7 @@ void use_keymap(char *name) /*{{{*/
 
 /*}}}*/
 
-char *what_keymap() /*{{{*/
+static char *what_keymap (void) /*{{{*/
 {
    return CBuf->keymap->name;
 }
@@ -951,8 +999,12 @@ char *what_keymap() /*{{{*/
 void jed_set_abort_char (int *c) /*{{{*/
 {
    char str[2];
-   int i;
    char ch = (char) *c;
+#if SLANG_VERSION >= 20000
+   SLkeymap_Type *km;
+#else
+      int i;
+#endif
 
    if (X_Set_Abort_Char_Hook != NULL)
      (void) (*X_Set_Abort_Char_Hook) ((unsigned char) ch);
@@ -960,13 +1012,22 @@ void jed_set_abort_char (int *c) /*{{{*/
    Jed_Abort_Char = (int) ch;
 
    str[0] = ch; str[1] = 0;
-   
+#if SLANG_VERSION < 20000   
    for (i = 0; i < SLANG_MAX_KEYMAPS; i++)
      {
 	if (SLKeyMap_List[i].keymap == NULL) continue;
 	SLang_undefine_key(str, &SLKeyMap_List[i]);
 	SLkm_define_key (str, (FVOID_STAR) kbd_quit, &SLKeyMap_List[i]);
      }
+#else
+   km = SLKeyMap_List_Root;
+   while (km != NULL)
+     {
+	SLang_undefine_key(str, km);
+	SLkm_define_key (str, (FVOID_STAR) kbd_quit, km);
+	km = km->next;
+     }
+#endif
 }
 
 /*}}}*/
@@ -974,8 +1035,7 @@ void jed_set_abort_char (int *c) /*{{{*/
 static SLKeymap_Function_Type *Flist_Context;
 static int Flist_Context_Len;
 
-#define MAX_USER_FLIST 100
-static char *Slang_Functions[MAX_USER_FLIST];
+static char *Slang_Functions[JED_MAX_ADD_COMPLETIONS];
 static char **Slang_Flist_Context;
 
 int next_function_list(char *buf) /*{{{*/
@@ -1005,7 +1065,7 @@ int next_function_list(char *buf) /*{{{*/
 	  }
      }
    
-   max = Slang_Functions + MAX_USER_FLIST;
+   max = Slang_Functions + JED_MAX_ADD_COMPLETIONS;
    while (Slang_Flist_Context < max)
      {
 	name = *Slang_Flist_Context;
@@ -1035,36 +1095,25 @@ int open_function_list(char *buf) /*{{{*/
 
 /*}}}*/
 
-void add_to_completion(char *name) /*{{{*/
+int add_to_completion(char *name) /*{{{*/
 {
-   char **p, *n;
    static char **last = Slang_Functions;
-   
+
+   if (last == Slang_Functions + JED_MAX_ADD_COMPLETIONS)
+     return 0;
+
    if (0 == SLang_is_defined (name))
      {
 	jed_verror ("add_completion: %s undefined", name);
-	return;
+	return -1;
      }
 
-   n = SLang_create_slstring (name);
-   if (n == NULL)
-     return;
+   name = SLang_create_slstring (name);
+   if (name == NULL)
+     return -1;
 
-   p = last;  /* Slang_Functions; */
-   
-   while (p < Slang_Functions + MAX_USER_FLIST)
-     {
-	if (*p == NULL)
-	  {
-	     *p = n;
-	     last = p + 1;
-	     return;
-	  }
-	p++;
-     }
-
-   SLang_free_slstring (n);
-   /* msg_error("Completion Quota Exceeded."); */
+   *last++ = name;
+   return 0;
 }
 
 /*}}}*/
@@ -1077,7 +1126,7 @@ int is_internal(char *f) /*{{{*/
 
 /*}}}*/
 
-void jed_copy_keymap_cmd (char *dest, char *src)
+static void copy_keymap_cmd (char *dest, char *src)
 {
    SLKeyMap_List_Type *ksrc;
 
@@ -1097,6 +1146,28 @@ void jed_copy_keymap_cmd (char *dest, char *src)
      jed_verror ("Unable to create keymap \"%s\"", dest);
 }
 
+void jed_call_cmd (char *str)
+{
+   SLKeyMap_List_Type *km;
+   int (*fp)(void);
+
+   if (*str == ' ')
+     {
+	jed_insert_string (str + 1);
+	return;
+     }
+   
+   if (*str == '@')
+     {
+	(void) do_macro_string (str + 1, 1);
+	return;
+     }
+   
+   km = CBuf->keymap;
+   if (NULL == (fp = (int (*)(void)) (SLang_find_key_function(str, km))))
+     jed_verror ("Internal function %s does not exist", str);
+   else (void) (*fp)();
+}
 
 typedef struct
 {
@@ -1120,13 +1191,17 @@ static int get_key_function_callback (void)
 }
 
    
-int jed_get_key_function (unsigned char *keystr, char **fun)
+static int push_key_binding (unsigned char *keystr)
 {
    SLang_Key_Type *key;
-   
+   char *fun;
+   int type;
+#ifdef SLKEY_F_SLANG
+   SLang_Name_Type *ref = NULL;
+#endif
+
    SLang_Key_TimeOut_Flag = 0;
 
-   *fun = NULL;
    if (keystr == NULL)
      {	
 	Key_Bufferp = Key_Buffer;
@@ -1135,9 +1210,9 @@ int jed_get_key_function (unsigned char *keystr, char **fun)
 	if (key == NULL)
 	  flush_input ();
 
-	if (SLKeyBoard_Quit && (SLang_Error == USER_BREAK))
+	if (SLKeyBoard_Quit || (SLang_get_error () == USER_BREAK))
 	  {
-	     SLang_Error = 0;
+	     SLang_set_error (0);
 	     SLKeyBoard_Quit = 0;
 	  }
      }
@@ -1149,48 +1224,302 @@ int jed_get_key_function (unsigned char *keystr, char **fun)
 	key = SLang_do_key (CBuf->keymap, get_key_function_callback);
      }
    
-   if (key == NULL)
-     return -1;
+   fun = NULL;
+   type = -1;
+
+   if (key != NULL)
+     switch (key->type)
+       {
+	case SLKEY_F_INTRINSIC:
+	  type = 1;
+	  fun = find_function_string (key->f.f);
+	  break;
+	  
+	case SLKEY_F_INTERPRET:
+	  fun = key->f.s;
+	  type = 0;
+	  if (fun[0] == '@') type = 2;
+	  else if (fun[0] == ' ') type = 3;
+	  break;
+	  
+#ifdef SLKEY_F_SLANG
+	case SLKEY_F_SLANG:
+	  type = 4;
+	  ref = key->f.slang_fun;
+	  break;
+#endif
+     }
    
-   if (key->type == SLKEY_F_INTRINSIC)
-     {
-	*fun = find_function_string (key->f.f);
-	return 1;
-     }
-   else 
-     {
-	*fun = key->f.s;
-	if (key->f.s[0] == '@') return 2;
-	if (key->f.s[0] == ' ') return 3;
-	return 0;
-     }
+   if (-1 == SLang_push_integer (type))
+     return -1;
+
+#ifdef SLKEY_F_SLANG
+   if (type == 4)
+     type = SLang_push_function (ref);
+   else
+#endif
+     type = SLang_push_string (fun);
+   
+   return type;
 }
 
 /*}}}*/
    
-   
-   
-     
-   
-void jed_call_cmd (char *str)
+static void get_key_binding (void)
 {
-   SLKeyMap_List_Type *km;
-   int (*fp)(void);
+   unsigned char *s = NULL;
 
-   if (*str == ' ')
+   if (SLang_Num_Function_Args == 1)
      {
-	jed_insert_string (str + 1);
-	return;
+	char *str;
+
+	if (-1 == SLang_pop_slstring (&str))
+	  return;
+
+	s = (unsigned char *) SLang_process_keystring (str);
+	SLang_free_slstring (str);
+
+	if (s == NULL)
+	  return;
      }
+
+   (void) push_key_binding (s);
+}
    
-   if (*str == '@')
+/* This is weird, Ultrix cc will not compile if set_key comes before unset_key */
+static void unset_key (char *key) /*{{{*/
+{
+   if (*key)
+     SLang_undefine_key(key, Global_Map);
+}
+
+/*}}}*/
+
+static void unset_key_in_keymap(char *key, char *map) /*{{{*/
+{
+   SLKeyMap_List_Type *kmap;
+
+   if (NULL == (kmap = SLang_find_keymap(map)))
      {
-	(void) do_macro_string (str + 1, 1);
+	jed_verror ("Unknown keymap: %s", map);
 	return;
      }
    
-   km = CBuf->keymap;
-   if (NULL == (fp = (int (*)(void)) (SLang_find_key_function(str, km))))
-     jed_verror ("Internal function %s does not exist", str);
-   else (void) (*fp)();
+   if (*key) SLang_undefine_key(key, kmap);
+}
+
+/*}}}*/
+
+static void set_key_in_keymap_1 (SLKeyMap_List_Type *kmap)
+{
+   char *key, *func;
+
+   if (-1 == SLang_pop_slstring (&key))
+     return;
+#ifdef SLKEY_F_SLANG
+   if (SLang_peek_at_stack () == SLANG_REF_TYPE)
+     {
+	SLang_Name_Type *nt = SLang_pop_function ();
+	if (nt == NULL)
+	  {
+	     SLang_free_slstring (key);
+	     return;
+	  }
+	if (-1 == SLkm_define_slkey (key, nt, kmap))
+	  {
+	     SLang_free_function (nt);
+	     SLang_free_slstring (key);
+	     return;
+	  }
+	SLang_free_slstring (key);
+	return;
+     }
+#endif
+   if (-1 == SLang_pop_slstring (&func))
+     {
+	SLang_free_slstring (key);
+	return;
+     }
+
+   if (*key)
+     (void) SLang_define_key(key, func, kmap);
+   
+   SLang_free_slstring (func);
+   SLang_free_slstring (key);
+}
+
+static void set_key_in_keymap (void)
+{
+   char *keymap;
+   SLKeyMap_List_Type *kmap;
+
+   if (-1 == SLang_pop_slstring (&keymap))
+     return;
+
+   if (NULL == (kmap = SLang_find_keymap(keymap)))
+     jed_verror ("Unknown keymap: %s", keymap);
+   else
+     set_key_in_keymap_1 (kmap);
+
+   SLang_free_slstring (keymap);
+}
+
+/*}}}*/
+
+static void set_key (void)
+{
+   set_key_in_keymap_1 (Global_Map);
+}
+
+
+/*}}}*/
+
+static int keymap_p(char *name) /*{{{*/
+{
+   return ! (NULL == SLang_find_keymap(name));
+}
+
+/*}}}*/
+
+
+static void make_keymap_cmd (char *km)
+{
+   copy_keymap_cmd (km, "global");
+}
+
+static void get_keymaps (void)
+{
+   int j, n;
+   char **keymaps;
+   SLang_Array_Type *at;
+#if SLANG_VERSION < 20000
+   unsigned int i;
+#else
+   SLkeymap_Type *km;
+#endif
+
+   n = 0;
+#if SLANG_VERSION < 20000
+   for (i = 0; i < SLANG_MAX_KEYMAPS; i++)
+     {
+	if (SLKeyMap_List[i].keymap == NULL)
+	  continue;
+
+	n++;
+     }
+#else
+   km = SLKeyMap_List_Root;
+   while (km != NULL)
+     {
+	n++;
+	km = km->next;
+     }
+#endif
+
+   if (NULL == (at = SLang_create_array (SLANG_STRING_TYPE, 0, NULL, &n, 1)))
+     return;
+
+   keymaps = (char **)at->data;
+   j = 0;
+#if SLANG_VERSION < 20000
+   for (i = 0; i < SLANG_MAX_KEYMAPS; i++)
+     {
+	if (SLKeyMap_List[i].keymap == NULL)
+	  continue;
+
+	if (j == n)
+	  break;		       /* keymap list changing?? */
+
+	if (NULL == (keymaps[j] = SLang_create_slstring (SLKeyMap_List[i].name)))
+	  {
+	     SLang_free_array (at);
+	     return;
+	  }
+	j++;
+     }
+#else
+   km = SLKeyMap_List_Root;
+   while (km != NULL)
+     {
+	if (j == n)
+	  break;		       /* keymap list changing?? */
+
+	if (NULL == (keymaps[j] = SLang_create_slstring (km->name)))
+	  {
+	     SLang_free_array (at);
+	     return;
+	  }
+	km = km->next;
+	j++;
+     }
+#endif
+   SLang_push_array (at, 1);
+}
+
+static void getkey_wchar_intrin (void)
+{
+   int ch;
+   SLwchar_Type wch;
+   
+   if (-1 == jed_getkey_wchar (&wch))
+     {
+	ch = (int) wch;
+	(void) SLang_push_integer (-ch);
+	return;
+     }
+   (void) SLclass_push_long_obj (SLANG_LONG_TYPE, (long)wch);
+}
+
+static void ungetkey_wchar_intrin (long *chp)
+{
+   if (*chp < 0)
+     {
+	int ch = -(int)*chp;
+	ungetkey (&ch);
+	return;
+     }
+   jed_ungetkey_wchar ((SLwchar_Type)*chp);
+}
+
+static SLang_Intrin_Fun_Type Keymap_Intrinsics [] = 
+{
+   MAKE_INTRINSIC_0("setkey", set_key, VOID_TYPE),
+   MAKE_INTRINSIC_0("definekey", set_key_in_keymap, VOID_TYPE),
+   MAKE_INTRINSIC_S("unsetkey", unset_key, VOID_TYPE),
+   MAKE_INTRINSIC("what_keymap", what_keymap, STRING_TYPE, 0),
+   MAKE_INTRINSIC_0("get_keymaps", get_keymaps, VOID_TYPE),
+   MAKE_INTRINSIC_S("keymap_p",  keymap_p, INT_TYPE),
+   MAKE_INTRINSIC("get_key_binding", get_key_binding, VOID_TYPE, 0),
+   MAKE_INTRINSIC_S("dump_bindings", dump_bindings, VOID_TYPE),
+   MAKE_INTRINSIC_SS("undefinekey", unset_key_in_keymap, VOID_TYPE),
+   MAKE_INTRINSIC_S("which_key", which_key, INT_TYPE),
+   MAKE_INTRINSIC_S("make_keymap", make_keymap_cmd, VOID_TYPE),
+   MAKE_INTRINSIC_SS("copy_keymap", copy_keymap_cmd, VOID_TYPE),
+   MAKE_INTRINSIC_S("use_keymap",  use_keymap, VOID_TYPE),
+   MAKE_INTRINSIC("_getkey", jed_getkey, INT_TYPE, 0),
+   MAKE_INTRINSIC_I("_ungetkey", ungetkey, VOID_TYPE),
+   MAKE_INTRINSIC_0("getkey", getkey_wchar_intrin, VOID_TYPE),
+   MAKE_INTRINSIC_1("ungetkey", ungetkey_wchar_intrin, VOID_TYPE, SLANG_LONG_TYPE),
+   SLANG_END_INTRIN_FUN_TABLE
+};
+
+static char *Key_Buffer_Ptr = Jed_Key_Buffer;
+static char *Last_Kbd_Command_Str_Ptr = Last_Kbd_Command_String;
+static char *Current_Kbd_Command_Str_Ptr = Current_Kbd_Command_String;
+
+static SLang_Intrin_Var_Type Keymap_Variables [] =
+{
+   MAKE_VARIABLE("LASTKEY", &Key_Buffer_Ptr, STRING_TYPE, 1),
+   MAKE_VARIABLE("LAST_KBD_COMMAND", &Last_Kbd_Command_Str_Ptr, STRING_TYPE, 1),
+   MAKE_VARIABLE("CURRENT_KBD_COMMAND", &Current_Kbd_Command_Str_Ptr, STRING_TYPE, 1),
+   MAKE_VARIABLE(NULL, NULL, 0, 0)
+};
+
+int jed_init_keymap_intrinsics (void)
+{
+   if ((-1 == SLadd_intrin_fun_table (Keymap_Intrinsics, NULL))
+       || (-1 == SLadd_intrin_var_table (Keymap_Variables, NULL)))
+     return -1;
+   
+   return 0;
 }
