@@ -1,5 +1,5 @@
 /* Drop-down menus */
-/* Copyright (c) 1999, 2000, 2002 John E. Davis
+/* Copyright (c) 1999, 2000, 2002, 2003, 2004, 2005, 2006 John E. Davis
  * This file is part of JED editor library source.
  *
  * You may distribute this file under the terms the GNU General Public
@@ -921,7 +921,7 @@ static int copy_menu (Menu_Popup_Type *dest, Menu_Node_Type *src)
 static void simulate_hline (unsigned int n)
 {
    while (n--)
-     SLsmg_write_nchars ("-", 1);
+     SLsmg_write_string ("-");
 }
 
 static void simulate_box (int r, int c, unsigned int dr, unsigned int dc)
@@ -933,21 +933,21 @@ static void simulate_box (int r, int c, unsigned int dr, unsigned int dc)
    dr--; 
    dc--;
    SLsmg_gotorc (r, c);
-   SLsmg_write_nchars ("+", 1);
+   SLsmg_write_string ("+");
    simulate_hline (dc-1);
-   SLsmg_write_nchars ("+", 1);
+   SLsmg_write_string ("+");
    SLsmg_gotorc (r + dr, c);
-   SLsmg_write_nchars ("+", 1);
+   SLsmg_write_string ("+");
    simulate_hline (dc-1);
-   SLsmg_write_nchars ("+", 1);
+   SLsmg_write_string ("+");
    
    rmax = r + dr;
    for (rr = r + 1; rr < rmax; rr++)
      {
 	SLsmg_gotorc (rr, c);
-	SLsmg_write_nchars ("|", 1);
+	SLsmg_write_string ("|");
 	SLsmg_gotorc (rr, c + dc);
-	SLsmg_write_nchars ("|", 1);
+	SLsmg_write_string ("|");
      }
 }
 
@@ -1627,7 +1627,11 @@ static int select_menu_cmd (void)
 	jed_exit_menu_bar ();
 	if (nt == NULL)
 	  return -1;
-	(void) SLang_push_anytype (((Menu_SLang_Fun_Type*)m)->client_data);
+	if ((-1 == SLang_start_arg_list ())
+	    || (-1 == SLang_push_anytype (((Menu_SLang_Fun_Type*)m)->client_data))
+	    || (-1 == SLang_end_arg_list ()))
+	  return -1;
+
 	return SLexecute_function (nt);
 
       case MENU_NODE_KEYSTRING:
@@ -1830,7 +1834,7 @@ SLang_Key_Type *jed_menu_do_key (void)
 
    ch = my_getkey ();
 
-   if (SLKeyBoard_Quit || SLang_Error)
+   if (SLKeyBoard_Quit || SLang_get_error ())
      {
 	jed_exit_menu_bar ();
 	return NULL;
@@ -1855,7 +1859,7 @@ static int exec_menubar_callback (Menu_Bar_Type *b, SLang_Name_Type **ntp)
    if (nt == NULL)
      return 0;
 
-   if (SLang_Error)
+   if (SLang_get_error ())
      return -1;
 
    if (-1 == SLang_push_string (b->name))
@@ -1885,11 +1889,11 @@ void jed_notify_menu_buffer_changed (void)
    if (NULL == (b = get_active_menubar ()))
      return;
 
-   err = SLang_Error;
-   SLang_Error = 0;
+   err = SLang_get_error ();
+   SLang_set_error (0);
    (void) exec_menubar_callback (b, &b->init_callback);
-   if (SLang_Error == 0)
-     SLang_Error = err;
+   if (SLang_get_error () == 0)
+     SLang_set_error (err);
 }
 
 
