@@ -1,5 +1,5 @@
 /* -*- mode: C; mode: fold; -*- */
-/* Copyright (c) 1992, 1998 John E. Davis
+/* Copyright (c) 1992, 1998, 2000, 2002, 2003, 2004, 2005, 2006 John E. Davis
  * This file is part of JED editor library source.
  *
  * You may distribute this file under the terms the GNU General Public
@@ -14,6 +14,7 @@
 #include "ins.h"
 #include "line.h"
 #include "misc.h"
+#include "screen.h"
 
 
 #define LAST_UNDO CBuf->undo->Last_Undo
@@ -142,15 +143,23 @@ int undo (void) /*{{{*/
    
    do
      {
+	int undo_type;
 #ifdef UNDO_HAS_REDO
 	line = (int) CURRENT_UNDO->linenum;
+	undo_type = CURRENT_UNDO->type & 0xFF;
 #else
 	line = (int) LAST_UNDO->linenum;
+	undo_type = LAST_UNDO->type & 0xFF;
 #endif
-	if ((line <= (int) CBuf->nup) || ((unsigned int) line > CBuf->nup + Max_LineNum))
+	if ((line <= (int) CBuf->nup) 
+	    || ((unsigned int) line >= CBuf->nup + Max_LineNum))
 	  {
-	     msg_error("Next undo lies outside visible buffer.");
-	     break;
+	     if (((unsigned int) line != CBuf->nup + Max_LineNum)
+		 || (undo_type == NLINSERT))
+	       {
+		  msg_error("Next undo lies outside visible buffer.");
+		  break;
+	       }
 	  }
 	line -= CBuf->nup;
 	goto_line(&line);
@@ -159,31 +168,27 @@ int undo (void) /*{{{*/
 #else
 	Point = LAST_UNDO->point;
 #endif
-
-#ifdef UNDO_HAS_REDO
-	switch (CURRENT_UNDO->type & 0xFF)
-#else
-	switch (LAST_UNDO->type & 0xFF)
-#endif
+		 
+	switch (undo_type)
 	  {
 #ifdef UNDO_HAS_REDO
-	   case CDELETE: ins_chars(CURRENT_UNDO->buf, CURRENT_UNDO->misc);
+	   case CDELETE: (void) jed_insert_nbytes (CURRENT_UNDO->buf, CURRENT_UNDO->misc);
 	     /* Point = CURRENT_UNDO->point; */
 #else
-	   case CDELETE: ins_chars(LAST_UNDO->buf, LAST_UNDO->misc);
+	   case CDELETE: (void) jed_insert_nbytes (LAST_UNDO->buf, LAST_UNDO->misc);
 	     /* Point = LAST_UNDO->point; */
 #endif
 	     break;
 
 #ifdef UNDO_HAS_REDO
-	   case CINSERT: deln(&CURRENT_UNDO->misc);
+	   case CINSERT: (void) jed_del_nbytes (CURRENT_UNDO->misc);
 #else
-	   case CINSERT: deln(&LAST_UNDO->misc);
+	   case CINSERT: (void) jed_del_nbytes (LAST_UNDO->misc);
 #endif
 	     break;
-	     
-	   case NLINSERT: del(); break;
-	     
+
+	   case NLINSERT: jed_del_wchar (); break;
+
 	   default: return(1);
 	  }
 
@@ -194,7 +199,7 @@ int undo (void) /*{{{*/
 	/*  above calls to ins_chars/deln/del may overflow the undo-ring */
 	if (CURRENT_UNDO->type & UNDO_UNCHANGED_FLAG) 
 	  {
-	     mark_buffer_modified (0);
+	     mark_buffer_modified (CBuf, 0, 1);
 	  }
 
 	if (CURRENT_UNDO == FIRST_UNDO) 
@@ -208,7 +213,7 @@ int undo (void) /*{{{*/
 #else
 	if (LAST_UNDO->type & UNDO_UNCHANGED_FLAG)
 	  {
-	     mark_buffer_modified (0);
+	     mark_buffer_modified (CBuf, 0, 1);
 	  }
 	
 	if (LAST_UNDO == FIRST_UNDO) LAST_UNDO->type = 0;
@@ -337,7 +342,7 @@ void mark_undo_boundary (Buffer *b) /*{{{*/
 
 void unmark_undo_boundary (Buffer *b) /*{{{*/
 {
-#if 1
+#if 0
    (void) b;
 #else
    Buffer *s = CBuf;
